"""

md_to_labelme.py

"Converts" a MegaDetector output .json file to labelme format (one .json per image
file).  "Convert" is in quotes because this is an opinionated transformation that 
requires a confidence threshold.

TODO:
   
* support variable confidence thresholds across classes
* support classification data

"""

#%% Imports and constants

import os
import json

from tqdm import tqdm

from multiprocessing.pool import Pool
from multiprocessing.pool import ThreadPool
from functools import partial

from md_visualization.visualization_utils import open_image
from md_utils.ct_utils import truncate_float
from detection.run_detector import DEFAULT_DETECTOR_LABEL_MAP

output_precision = 3
default_confidence_threshold = 0.15


#%% Functions

def get_labelme_dict_for_image(im,image_base_name=None,category_id_to_name=None,
                               info=None,confidence_threshold=None):
    """
    For the given image struct in MD results format, reformat the detections into
    labelme format.
    
    Args:
        im (dict): MegaDetector-formatted results dict, must include 'height' and 'width' fields
        image_base_name (str, optional): written directly to the 'imagePath' field in the output; 
            defaults to os.path.basename(im['file']).
        category_id_to_name (dict, optional): maps string-int category IDs to category names, defaults
            to the standard MD categories
        info (dict, optional): arbitrary metadata to write to the "detector_info" field in the output
            dict
        confidence_threshold (float, optional): only detections at or above this confidence threshold
            will be included in the output dict
    
    Return:
        dict: labelme-formatted dictionary, suitable for writing directly to a labelme-formatted .json file
    """
    
    if image_base_name is None:
        image_base_name = os.path.basename(im['file'])
        
    if category_id_to_name:
        category_id_to_name = DEFAULT_DETECTOR_LABEL_MAP
        
    if confidence_threshold is None:        
        confidence_threshold = -1.0
     
    output_dict = {}
    if info is not None:
        output_dict['detector_info'] = info
    output_dict['version'] = '5.3.0a0'
    output_dict['flags'] = {}
    output_dict['shapes'] = []
    output_dict['imagePath'] = image_base_name
    output_dict['imageHeight'] = im['height']
    output_dict['imageWidth'] = im['width']
    output_dict['imageData'] = None
    output_dict['detections'] = im['detections']
    
    # det = im['detections'][1]
    for det in im['detections']:
        
        if det['conf'] < confidence_threshold:
            continue
        
        shape = {}
        shape['conf'] = det['conf']
        shape['label'] = category_id_to_name[det['category']] 
        shape['shape_type'] = 'rectangle'
        shape['description'] = ''
        shape['group_id'] = None
        
        # MD boxes are [x_min, y_min, width_of_box, height_of_box] (relative)
        # 
        # labelme boxes are [[x0,y0],[x1,y1]] (absolute)
        x0 = truncate_float(det['bbox'][0] * im['width'],output_precision)
        y0 = truncate_float(det['bbox'][1] * im['height'],output_precision)
        x1 = truncate_float(x0 + det['bbox'][2] * im['width'],output_precision)
        y1 = truncate_float(y0 + det['bbox'][3] * im['height'],output_precision)
        shape['points'] = [[x0,y0],[x1,y1]]
        output_dict['shapes'].append(shape)
    
    # ...for each detection
    
    return output_dict

# ...def get_labelme_dict_for_image()


def _write_output_for_image(im,image_base,extension_prefix,info,
                            confidence_threshold,category_id_to_name,overwrite,
                            verbose=False):
    
    if 'failure' in im and im['failure'] is not None:
        assert 'detections' not in im or im['detections'] is None
        if verbose:
            print('Skipping labelme file generation for failed image {}'.format(
                im['file']))
        return
        
    im_full_path = os.path.join(image_base,im['file'])
    json_path = os.path.splitext(im_full_path)[0] + extension_prefix + '.json'
    
    if (not overwrite) and (os.path.isfile(json_path)):
        if verbose:
            print('Skipping existing file {}'.format(json_path))
        return

    output_dict = get_labelme_dict_for_image(im,
                                             image_base_name=os.path.basename(im_full_path),
                                             category_id_to_name=category_id_to_name,
                                             info=info,
                                             confidence_threshold=confidence_threshold)
            
    with open(json_path,'w') as f:
        json.dump(output_dict,f,indent=1)
    
# ...def write_output_for_image(...)



def md_to_labelme(results_file,image_base,confidence_threshold=None,
                  overwrite=False,extension_prefix='',n_workers=1,
                  use_threads=False,bypass_image_size_read=False,
                  verbose=False):
    """
    For all the images in [results_file], write a .json file in labelme format alongside the
    corresponding relative path within image_base.
    
    Args:
        results_file (str): MD results .json file to convert to Labelme format
        image_base (str): folder of images; filenames in [results_file] should be relative to
            this folder
        confidence_threshold (float, optional): only detections at or above this confidence threshold
            will be included in the output dict
        overwrite (bool, optional): whether to overwrite existing output files; if this is False
            and the output file for an image exists, we'll skip that image
        extension_prefix (str, optional): if non-empty, "extension_prefix" will be inserted before the .json 
            extension
        n_workers (int, optional): enables multiprocessing if > 1
        use_threads (bool, optional): if [n_workers] > 1, determines whether we parallelize via threads (True)
            or processes (False)
        bypass_image_size_read (bool, optional): if True, skips reading image sizes and trusts whatever is in
            the MD results file (don't set this to "True" if your MD results file doesn't contain image sizes)
        verbose (bool, optional): enables additionald ebug output    
    """
    
    if extension_prefix is None:
        extension_prefix = ''
        
    # Load MD results if necessary
    if isinstance(results_file,dict):
        md_results = results_file
    else:
        print('Loading MD results...')
        with open(results_file,'r') as f:
            md_results = json.load(f)
        
    # Read image sizes if necessary            
    if bypass_image_size_read:     
        
        print('Bypassing image size read')
        
    else:
    
        # TODO: parallelize this loop
    
        print('Reading image sizes...')
                
        # im = md_results['images'][0]
        for im in tqdm(md_results['images']):
            
            # Make sure this file exists
            im_full_path = os.path.join(image_base,im['file'])
            assert os.path.isfile(im_full_path), 'Image file {} does not exist'.format(im_full_path)
            
            json_path = os.path.splitext(im_full_path)[0] + extension_prefix + '.json'
            
            # Don't even bother reading sizes for files we're not going to generate
            if (not overwrite) and (os.path.isfile(json_path)):
                continue
            
            # Load w/h information if necessary
            if 'height' not in im or 'width' not in im:
                
                try:
                    pil_im = open_image(im_full_path)
                    im['width'] = pil_im.width
                    im['height'] = pil_im.height
                except Exception:
                    print('Warning: cannot open image {}, treating as a failure during inference'.format(
                        im_full_path))
                    if 'failure' not in im:
                        im['failure'] = 'Failure image access'        
    
            # ...if we need to read w/h information
            
        # ...for each image
        
    # ...if we're not bypassing image size read        
    
    print('\nGenerating labelme files...')
        
    # Write output
    if n_workers <= 1:
        for im in tqdm(md_results['images']):    
            _write_output_for_image(im,image_base,extension_prefix,md_results['info'],confidence_threshold,
                                   md_results['detection_categories'],overwrite,verbose)
    else:
        if use_threads:
            print('Starting parallel thread pool with {} workers'.format(n_workers))
            pool = ThreadPool(n_workers)
        else:
            print('Starting parallel process pool with {} workers'.format(n_workers))
            pool = Pool(n_workers)
        _ = list(tqdm(pool.imap(
                partial(_write_output_for_image,
                        image_base=image_base,extension_prefix=extension_prefix,
                        info=md_results['info'],confidence_threshold=confidence_threshold,
                        category_id_to_name=md_results['detection_categories'],
                        overwrite=overwrite,verbose=verbose),
                 md_results['images']),
                 total=len(md_results['images'])))
            
    # ...for each image
    
# ...def md_to_labelme()


#%% Interactive driver

if False:
    
    pass

    #%% Configure options
    
    md_results_file = os.path.expanduser('~/data/md-test.json')
    coco_output_file = os.path.expanduser('~/data/md-test-coco.json')
    image_folder = os.path.expanduser('~/data/md-test')    
    confidence_threshold = 0.2
    overwrite = True    
    
    
    #%% Programmatic execution
    
    md_to_labelme(results_file=md_results_file,
                  image_base=image_folder,
                  confidence_threshold=confidence_threshold,
                  overwrite=overwrite)

    
    #%% Command-line execution
    
    s = 'python md_to_labelme.py {} {} --confidence_threshold {}'.format(md_results_file,
                                                                         image_folder,
                                                                         confidence_threshold)
    if overwrite:
        s += ' --overwrite'
        
    print(s)
    import clipboard; clipboard.copy(s)


    #%% Opening labelme
    
    s = 'python labelme {}'.format(image_folder)
    print(s)
    import clipboard; clipboard.copy(s)
    

#%% Command-line driver

import sys,argparse

def main():

    parser = argparse.ArgumentParser(
        description='Convert MD output to labelme annotation format')
    parser.add_argument(
        'results_file',
        type=str,
        help='Path to MD results file (.json)')
    
    parser.add_argument(
        'image_base',
        type=str,
        help='Path to images (also the output folder)')
    
    parser.add_argument(
        '--confidence_threshold',
        type=float,
        default=default_confidence_threshold,
        help='Confidence threshold (default {})'.format(default_confidence_threshold)
        )
    
    parser.add_argument(
        '--overwrite',
        action='store_true',
        help='Overwrite existing labelme .json files')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    md_to_labelme(args.results_file,args.image_base,args.confidence_threshold,args.overwrite)

if __name__ == '__main__':
    main()



########
#
# video_utils.py
#
# Utilities for splitting, rendering, and assembling videos.
#
########

#%% Constants, imports, environment

import os
import cv2
import glob
import json

from collections import defaultdict
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from tqdm import tqdm
from typing import Container,Iterable,List
from functools import partial

from md_utils import path_utils
    
from md_visualization import visualization_utils as vis_utils


#%% Path utilities

VIDEO_EXTENSIONS = ('.mp4','.avi','.mpeg','.mpg')

def is_video_file(s: str, video_extensions: Container[str] = VIDEO_EXTENSIONS
                  ) -> bool:
    """
    Checks a file's extension against a hard-coded set of video file
    extensions.
    """
    
    ext = os.path.splitext(s)[1]
    return ext.lower() in video_extensions


def find_video_strings(strings: Iterable[str]) -> List[str]:
    """
    Given a list of strings that are potentially video file names, looks for
    strings that actually look like video file names (based on extension).
    """
    
    return [s for s in strings if is_video_file(s.lower())]


def find_videos(dirname: str, recursive: bool = False) -> List[str]:
    """
    Finds all files in a directory that look like video file names. Returns
    absolute paths.
    """
    
    if recursive:
        strings = glob.glob(os.path.join(dirname, '**', '*.*'), recursive=True)
    else:
        strings = glob.glob(os.path.join(dirname, '*.*'))
    return find_video_strings(strings)


#%% Function for rendering frames to video and vice-versa

# http://tsaith.github.io/combine-images-into-a-video-with-python-3-and-opencv-3.html

def frames_to_video(images, Fs, output_file_name, codec_spec='h264'):
    """
    Given a list of image files and a sample rate, concatenate those images into
    a video and write to [output_file_name].
    
    Note to self: h264 is a sensible default and generally works on Windows, but when this
    fails (which is around 50% of the time on Linux), I fall back to mp4v.
    """
    
    if codec_spec is None:
        codec_spec = 'h264'
        
    if len(images) == 0:
        return

    # Determine the width and height from the first image
    frame = cv2.imread(images[0])
    cv2.imshow('video',frame)
    height, width, channels = frame.shape

    # Define the codec and create VideoWriter object
    fourcc = cv2.VideoWriter_fourcc(*codec_spec)
    out = cv2.VideoWriter(output_file_name, fourcc, Fs, (width, height))

    for image in images:
        frame = cv2.imread(image)
        out.write(frame)

    out.release()
    cv2.destroyAllWindows()


def get_video_fs(input_video_file):
    """
    Get the frame rate of [input_video_file]
    """
    
    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)    
    vidcap = cv2.VideoCapture(input_video_file)
    Fs = vidcap.get(cv2.CAP_PROP_FPS)
    vidcap.release()
    return Fs


def frame_number_to_filename(frame_number):
    return 'frame{:06d}.jpg'.format(frame_number)


def video_to_frames(input_video_file, output_folder, overwrite=True, 
                    every_n_frames=None, verbose=False):
    """
    Render every frame of [input_video_file] to a .jpg in [output_folder]
    
    With help from:
        
    https://stackoverflow.com/questions/33311153/python-extracting-and-saving-video-frames
    """
    
    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)
    
    vidcap = cv2.VideoCapture(input_video_file)
    n_frames = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    Fs = vidcap.get(cv2.CAP_PROP_FPS)
    
    # If we're not over-writing, check whether all frame images already exist
    if overwrite == False:
        
        missing_frame_number = None
        frame_filenames = []
        
        for frame_number in range(0,n_frames):
            
            if every_n_frames is not None:
                if frame_number % every_n_frames != 0:
                    continue
            
            frame_filename = frame_number_to_filename(frame_number)
            frame_filename = os.path.join(output_folder,frame_filename)
            frame_filenames.append(frame_filename)
            if os.path.isfile(frame_filename):
                continue
            else:
                missing_frame_number = frame_number
                break
    
        # OpenCV seems to over-report the number of frames by 1 in some cases, or fails
        # to read the last frame; either way, I'm allowing one missing frame.
        allow_last_frame_missing = True
        
        if missing_frame_number is None or \
            (allow_last_frame_missing and (missing_frame_number == n_frames-1)):
            if verbose:
                print('Skipping video {}, all output frames exist'.format(input_video_file))
            return frame_filenames,Fs
        else:
            pass
            # print("Rendering video {}, couldn't find frame {}".format(
            #    input_video_file,missing_frame_number))
    
    # ...if we need to check whether to skip this video entirely
        
    if verbose:
        print('Reading {} frames at {} Hz from {}'.format(n_frames,Fs,input_video_file))

    frame_filenames = []

    # for frame_number in tqdm(range(0,n_frames)):
    for frame_number in range(0,n_frames):

        success,image = vidcap.read()
        if not success:
            assert image is None
            if verbose:
                print('Read terminating at frame {} of {}'.format(frame_number,n_frames))
            break

        if every_n_frames is not None:
            if frame_number % every_n_frames != 0:
                continue
            
        frame_filename = frame_number_to_filename(frame_number)
        frame_filename = os.path.join(output_folder,frame_filename)
        frame_filenames.append(frame_filename)
        
        if overwrite == False and os.path.isfile(frame_filename):
            # print('Skipping frame {}'.format(frame_filename))
            pass            
        else:
            try:
                if frame_filename.isascii():
                    cv2.imwrite(os.path.normpath(frame_filename),image)
                else:
                    is_success, im_buf_arr = cv2.imencode('.jpg', image)
                    im_buf_arr.tofile(frame_filename)
                assert os.path.isfile(frame_filename), \
                    'Output frame {} unavailable'.format(frame_filename)
            except KeyboardInterrupt:
                vidcap.release()
                raise
            except Exception as e:
                print('Error on frame {} of {}: {}'.format(frame_number,n_frames,str(e)))

    if verbose:
        print('\nExtracted {} of {} frames'.format(len(frame_filenames),n_frames))

    vidcap.release()    
    return frame_filenames,Fs


def _video_to_frames_for_folder(relative_fn,input_folder,output_folder_base,every_n_frames,overwrite,verbose):
    """
    Internal function to call video_to_frames in the context of video_folder_to_frames; 
    makes sure the right output folder exists, then calls video_to_frames.
    """    
    
    input_fn_absolute = os.path.join(input_folder,relative_fn)
    assert os.path.isfile(input_fn_absolute)

    # Create the target output folder
    output_folder_video = os.path.join(output_folder_base,relative_fn)
    os.makedirs(output_folder_video,exist_ok=True)

    # Render frames
    # input_video_file = input_fn_absolute; output_folder = output_folder_video
    frame_filenames,fs = video_to_frames(input_fn_absolute,output_folder_video,
                                         overwrite=overwrite,every_n_frames=every_n_frames,
                                         verbose=verbose)
    
    return frame_filenames,fs


def video_folder_to_frames(input_folder:str, output_folder_base:str, 
                           recursive:bool=True, overwrite:bool=True,
                           n_threads:int=1, every_n_frames:int=None,
                           verbose=False, parallelization_uses_threads=True):
    """
    For every video file in input_folder, create a folder within output_folder_base, and 
    render every frame of the video to .jpg in that folder.
    """
    
    # Recursively enumerate video files
    input_files_full_paths = find_videos(input_folder,recursive=recursive)
    print('Found {} videos in folder {}'.format(len(input_files_full_paths),input_folder))
    if len(input_files_full_paths) == 0:
        return [],[],[]
    
    input_files_relative_paths = [os.path.relpath(s,input_folder) for s in input_files_full_paths]
    input_files_relative_paths = [s.replace('\\','/') for s in input_files_relative_paths]
    
    os.makedirs(output_folder_base,exist_ok=True)    
    
    frame_filenames_by_video = []
    fs_by_video = []
    
    if n_threads == 1:
        # For each video
        #
        # input_fn_relative = input_files_relative_paths[0]
        for input_fn_relative in tqdm(input_files_relative_paths):
        
            frame_filenames,fs = \
                _video_to_frames_for_folder(input_fn_relative,input_folder,output_folder_base,
                                            every_n_frames,overwrite,verbose)
            frame_filenames_by_video.append(frame_filenames)
            fs_by_video.append(fs)
    else:
        if parallelization_uses_threads:
            print('Starting a worker pool with {} threads'.format(n_threads))
            pool = ThreadPool(n_threads)
        else:
            print('Starting a worker pool with {} processes'.format(n_threads))
            pool = Pool(n_threads)
        process_video_with_options = partial(_video_to_frames_for_folder, 
                                             input_folder=input_folder,
                                             output_folder_base=output_folder_base,
                                             every_n_frames=every_n_frames,
                                             overwrite=overwrite,
                                             verbose=verbose)
        results = list(tqdm(pool.imap(
            partial(process_video_with_options),input_files_relative_paths), 
                            total=len(input_files_relative_paths)))
        frame_filenames_by_video = [x[0] for x in results]
        fs_by_video = [x[1] for x in results]
        
    return frame_filenames_by_video,fs_by_video,input_files_full_paths
  

class FrameToVideoOptions:
    
    # zero-indexed
    nth_highest_confidence = 1
    
    
def frame_results_to_video_results(input_file,output_file,options:FrameToVideoOptions = None):
    """
    Given an API output file produced at the *frame* level, corresponding to a directory 
    created with video_folder_to_frames, map those frame-level results back to the 
    video level for use in Timelapse.
    
    Preserves everything in the input .json file other than the images.
    """

    if options is None:
        options = FrameToVideoOptions()
        
    # Load results
    with open(input_file,'r') as f:
        input_data = json.load(f)

    images = input_data['images']
    detection_categories = input_data['detection_categories']
    
    ## Break into videos
    
    video_to_frames = defaultdict(list) 
    
    # im = images[0]
    for im in tqdm(images):
        
        fn = im['file']
        video_name = os.path.dirname(fn)
        assert is_video_file(video_name)
        video_to_frames[video_name].append(im)
    
    print('Found {} unique videos in {} frame-level results'.format(
        len(video_to_frames),len(images)))
    
    output_images = []
    
    ## For each video...
    
    # video_name = list(video_to_frames.keys())[0]
    for video_name in tqdm(video_to_frames):
        
        frames = video_to_frames[video_name]
        
        all_detections_this_video = []
        
        # frame = frames[0]
        for frame in frames:
            if frame['detections'] is not None:
                all_detections_this_video.extend(frame['detections'])
            
        # At most one detection for each category for the whole video
        canonical_detections = []
            
        # category_id = list(detection_categories.keys())[0]
        for category_id in detection_categories:
            
            category_detections = [det for det in all_detections_this_video if \
                                   det['category'] == category_id]
            
            # Find the nth-highest-confidence video to choose a confidence value
            if len(category_detections) >= options.nth_highest_confidence:
                
                category_detections_by_confidence = sorted(category_detections, 
                                                           key = lambda i: i['conf'],reverse=True)
                canonical_detection = category_detections_by_confidence[options.nth_highest_confidence-1]
                canonical_detections.append(canonical_detection)
                                      
        # Prepare the output representation for this video
        im_out = {}
        im_out['file'] = video_name
        im_out['detections'] = canonical_detections
        
        # 'max_detection_conf' is no longer included in output files by default
        if False:
            im_out['max_detection_conf'] = 0
            if len(canonical_detections) > 0:
                confidences = [d['conf'] for d in canonical_detections]
                im_out['max_detection_conf'] = max(confidences)
        
        output_images.append(im_out)
        
    # ...for each video
    
    output_data = input_data
    output_data['images'] = output_images
    s = json.dumps(output_data,indent=1)
    
    # Write the output file
    with open(output_file,'w') as f:
        f.write(s)
    
        
#%% Test driver

if False:

    #%% Constants
    
    Fs = 30.01
    confidence_threshold = 0.75
    input_folder = 'z:\\'
    frame_folder_base = r'e:\video_test\frames'
    detected_frame_folder_base = r'e:\video_test\detected_frames'
    rendered_videos_folder_base = r'e:\video_test\rendered_videos'
    
    results_file = r'results.json'
    os.makedirs(detected_frame_folder_base,exist_ok=True)
    os.makedirs(rendered_videos_folder_base,exist_ok=True)
    
    
    #%% Split videos into frames
        
    frame_filenames_by_video,fs_by_video,video_filenames = \
        video_folder_to_frames(input_folder,frame_folder_base,recursive=True)
    
    
    #%% List image files, break into folders
    
    frame_files = path_utils.find_images(frame_folder_base,True)
    frame_files = [s.replace('\\','/') for s in frame_files]
    print('Enumerated {} total frames'.format(len(frame_files)))
    
    Fs = 30.01
    # Find unique folders
    folders = set()
    # fn = frame_files[0]
    for fn in frame_files:
        folders.add(os.path.dirname(fn))
    folders = [s.replace('\\','/') for s in folders]
    print('Found {} folders for {} files'.format(len(folders),len(frame_files)))
    
        
    #%% Load detector output
    
    with open(results_file,'r') as f:
        detection_results = json.load(f)
    detections = detection_results['images']
    detector_label_map = detection_results['detection_categories']
    for d in detections:
        d['file'] = d['file'].replace('\\','/').replace('video_frames/','')


    #%% Render detector frames
    
    # folder = list(folders)[0]
    for folder in folders:
        
        frame_files_this_folder = [fn for fn in frame_files if folder in fn]
        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        detection_results_this_folder = [d for d in detections if folder_relative in d['file']]
        print('Found {} detections in folder {}'.format(len(detection_results_this_folder),folder))
        assert len(frame_files_this_folder) == len(detection_results_this_folder)
        
        rendered_frame_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        os.makedirs(rendered_frame_output_folder,exist_ok=True)
        
        # d = detection_results_this_folder[0]
        for d in tqdm(detection_results_this_folder):
            
            input_file = os.path.join(frame_folder_base,d['file'])
            output_file = os.path.join(detected_frame_folder_base,d['file'])
            os.makedirs(os.path.dirname(output_file),exist_ok=True)
            vis_utils.draw_bounding_boxes_on_file(input_file,output_file,d['detections'],
                                                  confidence_threshold)
        
        # ...for each file in this folder
            
    # ...for each folder


    #%% Render output videos
            
    # folder = list(folders)[0]
    for folder in tqdm(folders):
        
        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        rendered_detector_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        assert os.path.isdir(rendered_detector_output_folder)
        
        frame_files_relative = os.listdir(rendered_detector_output_folder)
        frame_files_absolute = [os.path.join(rendered_detector_output_folder,s) \
                                for s in frame_files_relative]
        
        output_video_filename = os.path.join(rendered_videos_folder_base,folder_relative)
        os.makedirs(os.path.dirname(output_video_filename),exist_ok=True)
        
        original_video_filename = output_video_filename.replace(
            rendered_videos_folder_base,input_folder)
        assert os.path.isfile(original_video_filename)
        Fs = get_video_fs(original_video_filename)
                
        frames_to_video(frame_files_absolute, Fs, output_video_filename)

    # ...for each video



"""

ct_utils.py

Numeric/geometry/array utility functions.

"""

#%% Imports and constants

import inspect
import json
import math
import os

import jsonpickle
import numpy as np

# List of file extensions we'll consider images; comparisons will be case-insensitive
# (i.e., no need to include both .jpg and .JPG on this list).
image_extensions = ['.jpg', '.jpeg', '.gif', '.png']


#%% Functions

def truncate_float_array(xs, precision=3):
    """
    Vectorized version of truncate_float(...), truncates the fractional portion of each
    floating-point value to a specific number of floating-point digits.

    Args:
        xs (list): list of floats to truncate
        precision (int, optional): the number of significant digits to preserve, should be >= 1            
            
    Returns:
        list: list of truncated floats
    """

    return [truncate_float(x, precision=precision) for x in xs]


def truncate_float(x, precision=3):
    """
    Truncates the fractional portion of a floating-point value to a specific number of 
    floating-point digits.
    
    For example: 
        
        truncate_float(0.0003214884) --> 0.000321
        truncate_float(1.0003214884) --> 1.000321
    
    This function is primarily used to achieve a certain float representation
    before exporting to JSON.

    Args:
        x (float): scalar to truncate
        precision (int, optional): the number of significant digits to preserve, should be >= 1
        
    Returns:
        float: truncated version of [x]
    """

    assert precision > 0

    if np.isclose(x, 0):
        
        return 0
    
    elif (x > 1):
        
        fractional_component = x - 1.0
        return 1 + truncate_float(fractional_component)
    
    else:
        
        # Determine the factor, which shifts the decimal point of x
        # just behind the last significant digit.
        factor = math.pow(10, precision - 1 - math.floor(math.log10(abs(x))))
        
        # Shift decimal point by multiplication with factor, flooring, and
        # division by factor.
        return math.floor(x * factor)/factor


def args_to_object(args, obj):
    """
    Copies all fields from a Namespace (typically the output from parse_args) to an
    object. Skips fields starting with _. Does not check existence in the target
    object.

    Args:
        args (argparse.Namespace): the namespace to convert to an object
        obj (object): object whose whose attributes will be updated
        
    Returns:
        object: the modified object (modified in place, but also returned)
    """
    
    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)

    return obj


def pretty_print_object(obj, b_print=True):
    """
    Converts an arbitrary object to .json, optionally printing the .json representation.
    
    Args:
        obj (object): object to print
        b_print (bool, optional): whether to print the object
        
    Returns:
        str: .json reprepresentation of [obj]
    """

    # _ = pretty_print_object(obj)

    # TODO: it's sloppy that I'm making a module-wide change here.
    jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)
    a = jsonpickle.encode(obj)
    s = '{}'.format(a)
    if b_print:
        print(s)
    return s


def is_list_sorted(L, reverse=False):
    """
    Returns True if the list L appears to be sorted, otherwise False.
    
    Calling is_list_sorted(L,reverse=True) is the same as calling
    is_list_sorted(L.reverse(),reverse=False).
    
    Args:
        L (list): list to evaluate
        reverse (bool, optional): whether to reverse the list before evaluating sort status 
    
    Returns:
        bool: True if the list L appears to be sorted, otherwise False
    """
    
    if reverse:
        return all(L[i] >= L[i + 1] for i in range(len(L)-1))
    else:
        return all(L[i] <= L[i + 1] for i in range(len(L)-1))
        

def write_json(path, content, indent=1):
    """
    Standardized wrapper for json.dump().
    
    Args:
        path (str): filename to write to
        content (object): object to dump
        indent (int, optional): indentation depth passed to json.dump
    """
    
    with open(path, 'w') as f:
        json.dump(content, f, indent=indent)


def convert_yolo_to_xywh(yolo_box):
    """
    Converts a YOLO format bounding box to [x_min, y_min, width_of_box, height_of_box].

    Args:
        yolo_box (list): bounding box of format [x_center, y_center, width_of_box, height_of_box]

    Returns:
        list: bbox with coordinates represented as [x_min, y_min, width_of_box, height_of_box]
    """
    
    x_center, y_center, width_of_box, height_of_box = yolo_box
    x_min = x_center - width_of_box / 2.0
    y_min = y_center - height_of_box / 2.0
    return [x_min, y_min, width_of_box, height_of_box]


def convert_xywh_to_tf(api_box):
    """
    Converts an xywh bounding box (the format used in MD output) to the [y_min, x_min, y_max, x_max] 
    format that the TensorFlow Object Detection API uses.

    Args:
        api_box: bbox output by the batch processing API [x_min, y_min, width_of_box, height_of_box]

    Returns:
        list: bbox with coordinates represented as [y_min, x_min, y_max, x_max]
    """
    
    x_min, y_min, width_of_box, height_of_box = api_box
    x_max = x_min + width_of_box
    y_max = y_min + height_of_box
    return [y_min, x_min, y_max, x_max]


def convert_xywh_to_xyxy(api_bbox):
    """
    Converts an xywh bounding box (the MD output format) to an xyxy bounding box.

    Args:
        api_bbox (list): bbox formatted as [x_min, y_min, width_of_box, height_of_box]

    Returns:
        list: bbox formatted as [x_min, y_min, x_max, y_max]
    """

    x_min, y_min, width_of_box, height_of_box = api_bbox
    x_max, y_max = x_min + width_of_box, y_min + height_of_box
    return [x_min, y_min, x_max, y_max]


def get_iou(bb1, bb2):
    """
    Calculates the intersection over union (IoU) of two bounding boxes.

    Adapted from:
        
    https://stackoverflow.com/questions/25349178/calculating-percentage-of-bounding-box-overlap-for-image-detector-evaluation

    Args:
        bb1 (list): [x_min, y_min, width_of_box, height_of_box]
        bb2 (list): [x_min, y_min, width_of_box, height_of_box]

    Returns:
        float: intersection_over_union, a float in [0, 1]
    """

    bb1 = convert_xywh_to_xyxy(bb1)
    bb2 = convert_xywh_to_xyxy(bb2)

    assert bb1[0] < bb1[2], 'Malformed bounding box (x2 >= x1)'
    assert bb1[1] < bb1[3], 'Malformed bounding box (y2 >= y1)'

    assert bb2[0] < bb2[2], 'Malformed bounding box (x2 >= x1)'
    assert bb2[1] < bb2[3], 'Malformed bounding box (y2 >= y1)'

    # Determine the coordinates of the intersection rectangle
    x_left = max(bb1[0], bb2[0])
    y_top = max(bb1[1], bb2[1])
    x_right = min(bb1[2], bb2[2])
    y_bottom = min(bb1[3], bb2[3])

    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes is always an
    # axis-aligned bounding box
    intersection_area = (x_right - x_left) * (y_bottom - y_top)

    # Compute the area of both AABBs
    bb1_area = (bb1[2] - bb1[0]) * (bb1[3] - bb1[1])
    bb2_area = (bb2[2] - bb2[0]) * (bb2[3] - bb2[1])

    # Compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the intersection area.
    iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
    assert iou >= 0.0, 'Illegal IOU < 0'
    assert iou <= 1.0, 'Illegal IOU > 1'
    return iou


def _get_max_conf_from_detections(detections):
    """
    Internal function used by get_max_conf(); don't call this directly.
    """
    
    max_conf = 0.0
    if detections is not None and len(detections) > 0:
        confidences = [det['conf'] for det in detections]
        max_conf = max(confidences)
    return max_conf


def get_max_conf(im):
    """
    Given an image dict in the MD output format, computes the maximum detection confidence for any 
    class.  Returns 0.0 (rather than None) if there was a failure or 'detections' isn't present.
    
    Args:
        im (dict): image dictionary in the MD output format (with a 'detections' field)
        
    Returns:
        float: the maximum detection confidence across all classes
    """
    
    max_conf = 0.0
    if 'detections' in im and im['detections'] is not None and len(im['detections']) > 0:
        max_conf = _get_max_conf_from_detections(im['detections'])
    return max_conf


def point_dist(p1,p2):
    """
    Computes the distance between two points, represented as length-two tuples.
    
    Args:
        p1: point, formatted as (x,y)
        p2: point, formatted as (x,y)
        
    Returns:
        float: the Euclidean distance between p1 and p2
    """
    
    return math.sqrt( ((p1[0]-p2[0])**2) + ((p1[1]-p2[1])**2) )


def rect_distance(r1, r2, format='x0y0x1y1'):
    """
    Computes the minimum distance between two axis-aligned rectangles, each represented as 
    (x0,y0,x1,y1) by default.
    
    Can also specify "format" as x0y0wh for MD-style bbox formatting (x0,y0,w,h).
    
    Args:
        r1: rectangle, formatted as (x0,y0,x1,y1) or (x0,y0,xy,y1)
        r2: rectangle, formatted as (x0,y0,x1,y1) or (x0,y0,xy,y1)
        format (str, optional): whether the boxes are formatted as 'x0y0x1y1' (default) or 'x0y0wh'
        
    Returns:
        float: the minimum distance between r1 and r2
    """
    
    assert format in ('x0y0x1y1','x0y0wh'), 'Illegal rectangle format {}'.format(format)
    
    if format == 'x0y0wh':
        # Convert to x0y0x1y1 without modifying the original rectangles
        r1 = [r1[0],r1[1],r1[0]+r1[2],r1[1]+r1[3]]
        r2 = [r2[0],r2[1],r2[0]+r2[2],r2[1]+r2[3]]
        
    # https://stackoverflow.com/a/26178015
    x1, y1, x1b, y1b = r1
    x2, y2, x2b, y2b = r2
    left = x2b < x1
    right = x1b < x2
    bottom = y2b < y1
    top = y1b < y2
    if top and left:
        return point_dist((x1, y1b), (x2b, y2))
    elif left and bottom:
        return point_dist((x1, y1), (x2b, y2b))
    elif bottom and right:
        return point_dist((x1b, y1), (x2, y2b))
    elif right and top:
        return point_dist((x1b, y1b), (x2, y2))
    elif left:
        return x1 - x2b
    elif right:
        return x2 - x1b
    elif bottom:
        return y1 - y2b
    elif top:
        return y2 - y1b
    else:
        return 0.0


def split_list_into_fixed_size_chunks(L,n):
    """
    Split the list or tuple L into chunks of size n (allowing chunks of size n-1 if necessary,
    i.e. len(L) does not have to be a multiple of n).
    
    Args:
        L (list): list to split into chunks
        n (int): preferred chunk size
        
    Returns:
        list: list of chunks, where each chunk is a list of length n or n-1
    """
    
    return [L[i * n:(i + 1) * n] for i in range((len(L) + n - 1) // n )]


def split_list_into_n_chunks(L, n, chunk_strategy='greedy'):
    """
    Splits the list or tuple L into n equally-sized chunks (some chunks may be one 
    element smaller than others, i.e. len(L) does not have to be a multiple of n).
    
    chunk_strategy can be "greedy" (default, if there are k samples per chunk, the first
    k go into the first chunk) or "balanced" (alternate between chunks when pulling
    items from the list).
                                              
    Args:
        L (list): list to split into chunks
        n (int): number of chunks
        chunk_strategy (str, optiopnal): "greedy" or "balanced"; see above
        
    Returns:
        list: list of chunks, each of which is a list
    """
    
    if chunk_strategy == 'greedy':
        k, m = divmod(len(L), n)
        return list(L[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))
    elif chunk_strategy == 'balanced':
        chunks = [ [] for _ in range(n) ]
        for i_item,item in enumerate(L):
            i_chunk = i_item % n
            chunks[i_chunk].append(item)
        return chunks
    else:
        raise ValueError('Invalid chunk strategy: {}'.format(chunk_strategy))


def sort_dictionary_by_key(d,reverse=False):
    """
    Sorts the dictionary [d] by key.
    
    Args:
        d (dict): dictionary to sort
        reverse (bool, optional): whether to sort in reverse (descending) order
        
    Returns:
        dict: sorted copy of [d]
    """
    
    d = dict(sorted(d.items(),reverse=reverse))
    return d
    

def sort_dictionary_by_value(d,sort_values=None,reverse=False):
    """
    Sorts the dictionary [d] by value.  If sort_values is None, uses d.values(),
    otherwise uses the dictionary sort_values as the sorting criterion.
    
    Args:
        d (dict): dictionary to sort
        sort_values (dict, optional): dictionary mapping keys in [d] to sort values (defaults 
            to None, uses [d] itself for sorting)
        reverse (bool, optional): whether to sort in reverse (descending) order
    
    Returns:
        dict: sorted copy of [d]
    """
    
    if sort_values is None:
        d = {k: v for k, v in sorted(d.items(), key=lambda item: item[1], reverse=reverse)}
    else:
        d = {k: v for k, v in sorted(d.items(), key=lambda item: sort_values[item[0]], reverse=reverse)}
    return d


def invert_dictionary(d):
    """
    Creates a new dictionary that maps d.values() to d.keys().  Does not check
    uniqueness.
    
    Args:
        d (dict): dictionary to invert
    
    Returns:
        dict: inverted copy of [d]
    """
    
    return {v: k for k, v in d.items()}


def image_file_to_camera_folder(image_fn):
    r"""
    Removes common overflow folders (e.g. RECNX101, RECNX102) from paths, i.e. turn:
        
    a\b\c\RECNX101\image001.jpg
    
    ...into:
        
    a\b\c

    Returns the same thing as os.dirname() (i.e., just the folder name) if no overflow folders are 
    present.

    Always converts backslashes to slashes.
    
    Args:
        image_fn (str): the image filename from which we should remove overflow folders
        
    Returns:
        str: a version of [image_fn] from which camera overflow folders have been removed
    """
    
    import re
    
    # 100RECNX is the overflow folder style for Reconyx cameras
    # 100EK113 is (for some reason) the overflow folder style for Bushnell cameras
    # 100_BTCF is the overflow folder style for Browning cameras
    # 100MEDIA is the overflow folder style used on a number of consumer-grade cameras
    patterns = ['\/\d+RECNX\/','\/\d+EK\d+\/','\/\d+_BTCF\/','\/\d+MEDIA\/']
    
    image_fn = image_fn.replace('\\','/')    
    for pat in patterns:
        image_fn = re.sub(pat,'/',image_fn)
    camera_folder = os.path.dirname(image_fn)
    
    return camera_folder
    

def is_float(v):
    """
    Determines whether v is either a float or a string representation of a float.
    
    Args:
        v (object): object to evaluate
        
    Returns:
        bool: True if [v] is a float or a string representation of a float, otherwise False
    """
    
    try:
        _ = float(v)
        return True
    except ValueError:
        return False


def is_iterable(x):
    """
    Uses duck typing to assess whether [x] is iterable (list, set, dict, etc.).
    
    Args:
        x (object): the object to test
    
    Returns:
        bool: True if [x] appears to be iterable, otherwise False
    """
    
    try:
        _ = iter(x)
    except:
       return False
    return True


def is_empty(v):
    """
    A common definition of "empty" used throughout the repo, particularly when loading
    data from .csv files.  "empty" includes None, '', and NaN.
    
    Args:
        v: the object to evaluate for emptiness
        
    Returns:
        bool: True if [v] is None, '', or NaN, otherwise False
    """
    if v is None:
        return True
    if isinstance(v,str) and v == '':
        return True
    if isinstance(v,float) and np.isnan(v):
        return True
    return False


def isnan(v):
    """
    Returns True if v is a nan-valued float, otherwise returns False.
    
    Args:
        v: the object to evaluate for nan-ness
    
    Returns:
        bool: True if v is a nan-valued float, otherwise False
    """
    
    try:        
        return np.isnan(v)
    except Exception:
        return False


def sets_overlap(set1, set2):
    """
    Determines whether two sets overlap.
    
    Args:
        set1 (set): the first set to compare (converted to a set if it's not already)
        set2 (set): the second set to compare (converted to a set if it's not already)
        
    Returns:
        bool: True if any elements are shared between set1 and set2
    """
    
    return not set(set1).isdisjoint(set(set2))



#%% Test drivers

if False:
    
    pass
    
    #%% Test image_file_to_camera_folder()
    
    relative_path = 'a/b/c/d/100EK113/blah.jpg'
    print(image_file_to_camera_folder(relative_path))
    
    relative_path = 'a/b/c/d/100RECNX/blah.jpg'
    print(image_file_to_camera_folder(relative_path))
    
    
    #%% Test a few rectangle distances
    
    r1 = [0,0,1,1]; r2 = [0,0,1,1]; assert rect_distance(r1,r2)==0
    r1 = [0,0,1,1]; r2 = [0,0,1,100]; assert rect_distance(r1,r2)==0
    r1 = [0,0,1,1]; r2 = [1,1,2,2]; assert rect_distance(r1,r2)==0
    r1 = [0,0,1,1]; r2 = [1.1,0,0,1.1]; assert abs(rect_distance(r1,r2)-.1) < 0.00001
    
    r1 = [0.4,0.8,10,22]; r2 = [100, 101, 200, 210.4]; assert abs(rect_distance(r1,r2)-119.753) < 0.001
    r1 = [0.4,0.8,10,22]; r2 = [101, 101, 200, 210.4]; assert abs(rect_distance(r1,r2)-120.507) < 0.001    
    r1 = [0.4,0.8,10,22]; r2 = [120, 120, 200, 210.4]; assert abs(rect_distance(r1,r2)-147.323) < 0.001



"""

cct_json_utils.py

Utilities for working with COCO Camera Traps .json databases

https://github.com/agentmorris/MegaDetector/blob/master/data_management/README.md#coco-cameratraps-format

"""

#%% Constants and imports

import json
import os

from tqdm import tqdm
from collections import defaultdict, OrderedDict


#%% Classes

class CameraTrapJsonUtils:
    """
    Miscellaneous utility functions for working with COCO Camera Traps databases
    """
    
    @staticmethod
    def annotations_to_string(annotations, cat_id_to_name):
        """
        Given a list of annotations and a mapping from class IDs to names, produces
        a comma-delimited string containing a list of class names, sorted alphabetically.
        
        Args:
            annotations (list): a list of annotation dicts
            cat_id_to_name (dict): a dict mapping category IDs to category names
            
        Returns:
            str: a comma-delimited list of class names
        """
        
        class_names = CameraTrapJsonUtils.annotations_to_class_names(annotations, cat_id_to_name)
        return ','.join(class_names)


    @staticmethod
    def annotations_to_class_names(annotations, cat_id_to_name):
        """
        Given a list of annotations and a mapping from class IDs to names, produces
        a list of class names, sorted alphabetically.
        
        Args:
            annotations (list): a list of annotation dicts
            cat_id_to_name (dict): a dict mapping category IDs to category names
            
        Returns:
            list: a list of class names present in [annotations]
        """
        
        # Collect all names
        class_names = [cat_id_to_name[ann['category_id']] for ann in annotations]
        # Make names unique and sort
        class_names = sorted(set(class_names))
        return class_names


    @staticmethod
    def order_db_keys(db):
        """
        Given a dict representing a JSON database in the COCO Camera Trap
        format, returns an OrderedDict with keys in the order of 'info',
        'categories', 'annotations' and 'images'. When this OrderedDict is
        serialized with json.dump(), the order of the keys are preserved.

        Args:
            db (dict): a JSON database in the COCO Camera Trap format

        Returns:
            dict: the same content as [db] but as an OrderedDict with keys ordered for 
                readability
        """
        
        ordered = OrderedDict([
            ('info', db['info']),
            ('categories', db['categories']),
            ('annotations', db['annotations']),
            ('images', db['images'])])
        return ordered


    @staticmethod
    def group_annotations_by_image_field(db_indexed, image_field='seq_id'):
        """
        Given an instance of IndexedJsonDb, group annotation entries by a field in the
        image entry.  Typically used to find all the annotations associated with a sequence.
        
        Args:
            db_indexed (IndexedJsonDb): an initialized IndexedJsonDb, typically loaded from a 
                COCO Camera Traps .json file
            image_field (str, optional): a field by which to group annotations (defaults
                to 'seq_id')
        
        Returns:
            dict: a dict mapping objects (typically strings, in fact typically sequence IDs) to
                lists of annotations
        """
        
        image_id_to_image_field = {}
        for image_id, image_entry in db_indexed.image_id_to_image.items():
            image_id_to_image_field[image_id] = image_entry[image_field]

        res = defaultdict(list)
        for annotations in db_indexed.image_id_to_annotations.values():
            for annotation_entry in annotations:
                field_value = image_id_to_image_field[annotation_entry['image_id']]
                res[field_value].append(annotation_entry)
        return res


    @staticmethod
    def get_entries_for_locations(db, locations):
        """
        Given a dict representing a JSON database in the COCO Camera Trap format, returns a dict
        with the 'images' and 'annotations' fields in the CCT format, each is an array that only
        includes entries in the original [db] that are in the [locations] set.
        
        Args:
            db (dict): a dict representing a JSON database in the COCO Camera Trap format
            locations (set): a set or list of locations to include; each item is a string

        Returns:
            dict: a dict with the 'images' and 'annotations' fields in the CCT format
        """
        
        locations = set(locations)
        print('Original DB has {} image and {} annotation entries.'.format(
            len(db['images']), len(db['annotations'])))
        new_db = { 'images': [], 'annotations': [] }
        new_images = set()
        for i in db['images']:
            # cast location to string as the entries in locations are strings
            if str(i['location']) in locations:
                new_db['images'].append(i)
                new_images.add(i['id'])
        for a in db['annotations']:
            if a['image_id'] in new_images:
                new_db['annotations'].append(a)
        print(
            'New DB has {} image and {} annotation entries.'.format(
                len(new_db['images']), len(new_db['annotations'])))
        return new_db


class IndexedJsonDb:
    """
    Wrapper for a COCO Camera Traps database.

    Handles boilerplate dictionary creation that we do almost every time we load
    a .json database.
    """

    def __init__(self, 
                 json_filename,
                 b_normalize_paths=False,
                 filename_replacements=None,
                 b_convert_classes_to_lower=True,
                 b_force_forward_slashes=True):
        """
        Constructor for IndexedJsonDb that loads from a .json file or CCT-formatted dict.
        
        Args:
            json_filename (str): filename to load, or an already-loaded dict
            b_normalize_paths (bool, optional): whether to invoke os.path.normpath on 
                all filenames.  Not relevant if b_force_forward_slashes is True.
            filename_replacements (dict, optional): a set of string --> string mappings
                that will trigger replacements in all filenames, typically used to remove
                leading folders
            b_convert_classes_to_lower (bool, optional): whether to convert all class
                names to lowercase
            b_force_forward_slashes (bool, optional): whether to convert backslashes to
                forward slashes in all path names            
        """
        
        if isinstance(json_filename, str):
            with open(json_filename) as f:
                self.db = json.load(f)
        else:
            self.db = json_filename

        assert 'images' in self.db, (
            f'Could not find image list in file {json_filename}, are you sure '
            'this is a COCO camera traps file?')
        
        if b_convert_classes_to_lower:
            # Convert classnames to lowercase to simplify comparisons later
            for c in self.db['categories']:
                c['name'] = c['name'].lower()

        # Normalize paths to simplify comparisons later
        if b_normalize_paths:
            for im in self.db['images']:
                im['file_name'] = os.path.normpath(im['file_name'])

        if b_force_forward_slashes:
            for im in self.db['images']:
                im['file_name'] = im['file_name'].replace('\\','/')

        if filename_replacements is not None:
            for s in filename_replacements:
                # Make custom replacements in filenames, typically used to
                # accommodate changes in root paths after DB construction
                r = filename_replacements[s]
                for im in self.db['images']:
                    im['file_name'] = im['file_name'].replace(s, r)

        ### Build useful mappings to facilitate working with the DB

        # Category ID <--> name
        self.cat_id_to_name = {
            cat['id']: cat['name'] for cat in self.db['categories']}
        self.cat_name_to_id = {
            cat['name']: cat['id'] for cat in self.db['categories']}

        # Image filename --> ID
        self.filename_to_id = {
            im['file_name']: im['id'] for im in self.db['images']}

        # Image ID --> image object
        self.image_id_to_image = {im['id']: im for im in self.db['images']}

        # Image ID --> annotations
        # Each image can potentially multiple annotations, hence using lists
        self.image_id_to_annotations = {}
        self.image_id_to_annotations = defaultdict(list)
        for ann in self.db['annotations']:
            self.image_id_to_annotations[ann['image_id']].append(ann)

    # ...__init__


    def get_annotations_for_image(self, image):
        """
        Finds all the annnotations associated with the image dict [image].
        
        Args:
            image (dict): an image dict loaded from a CCT .json file.  Only the 'id' field
                is used.
            
        Returns:
            list: list of annotations associated with this image.  Returns None if the db 
                has not been loaded, or [] if no annotations are available for this image.
        """
        
        if self.db is None:
            return None

        if image['id'] not in self.image_id_to_annotations:
            return []

        image_annotations = self.image_id_to_annotations[image['id']]
        return image_annotations


    def get_classes_for_image(self, image):
        """
        Returns a list of class names associated with [image].

        Args:
            image (dict): an image dict loaded from a CCT .json file.  Only the 'id' field
                is used.
            
        Returns:
            list: list of class names associated with this image.  Returns None if the db 
                has not been loaded, or [] if no annotations are available for this image.
        """
        
        if self.db is None:
            return None

        if image['id'] not in self.image_id_to_annotations:
            return []

        class_ids = []
        image_annotations = self.image_id_to_annotations[image['id']]
        for ann in image_annotations:
            class_ids.append(ann['category_id'])
        class_ids = sorted(set(class_ids))
        class_names = [self.cat_id_to_name[x] for x in class_ids]

        return class_names

# ...class IndexedJsonDb

class SequenceOptions:
    """
    Options parameterizing the grouping of images into sequences by time.
    """
    
    #: Images separated by <= this duration will be grouped into the same sequence.
    episode_interval_seconds = 60.0

    
#%% Functions

def create_sequences(image_info,options=None):
    """
    Synthesizes episodes/sequences/bursts for the images in [image_info].
    
    Modifies [image_info], populating the 'seq_id', 'seq_num_frames', and 'frame_num' fields
    for each image.
    
    Args:
        image_info (dict): a list of dicts in CCT format, i.e. with fields 'file_name' (str),
            'datetime' (datetime), and 'location' (str).    
    """
    
    if options is None:
        options = SequenceOptions()
        
    # Find all unique locations
    locations = set()
    for im in image_info:
        locations.add(im['location'])
        
    print('Found {} locations'.format(len(locations)))    
    locations = list(locations)
    locations.sort()
    
    all_sequences = set()
    
    # i_location = 0; location = locations[i_location]
    for i_location,location in tqdm(enumerate(locations),total=len(locations)):
        
        images_this_location = [im for im in image_info if im['location'] == location]    
        
        # Sorting datetimes fails when there are None's in the list.  So instead of sorting datetimes 
        # directly, sort tuples with a boolean for none-ness, then the datetime itself.
        #
        # https://stackoverflow.com/questions/18411560/sort-list-while-pushing-none-values-to-the-end
        sorted_images_this_location = sorted(images_this_location, 
                                             key = lambda im: (im['datetime'] is None,im['datetime']))
        
        sequence_id_to_images_this_location = defaultdict(list)

        current_sequence_id = None
        next_frame_number = 0
        next_sequence_number = 0
        previous_datetime = None
            
        # previous_datetime = sorted_images_this_location[0]['datetime']
        # im = sorted_images_this_location[1]
        for im in sorted_images_this_location:
            
            invalid_datetime = False
            
            if previous_datetime is None:
                delta = None
            elif im['datetime'] is None:
                invalid_datetime = True
            else:
                delta = (im['datetime'] - previous_datetime).total_seconds()
            
            # Start a new sequence if necessary, including the case where this datetime is invalid
            if delta is None or delta > options.episode_interval_seconds or invalid_datetime:
                next_frame_number = 0
                current_sequence_id = 'location_{}_sequence_index_{}'.format(
                    location,str(next_sequence_number).zfill(5))
                next_sequence_number = next_sequence_number + 1
                assert current_sequence_id not in all_sequences
                all_sequences.add(current_sequence_id)                
                
            im['seq_id'] = current_sequence_id
            im['seq_num_frames'] = None
            im['frame_num'] = next_frame_number
            sequence_id_to_images_this_location[current_sequence_id].append(im)
            next_frame_number = next_frame_number + 1
            
            # If this was an invalid datetime, this will record the previous datetime
            # as None, which will force the next image to start a new sequence.
            previous_datetime = im['datetime']
        
        # ...for each image in this location
    
        # Fill in seq_num_frames
        for seq_id in sequence_id_to_images_this_location.keys():
            assert seq_id in sequence_id_to_images_this_location
            images_this_sequence = sequence_id_to_images_this_location[seq_id]
            assert len(images_this_sequence) > 0
            for im in images_this_sequence:
                im['seq_num_frames'] = len(images_this_sequence)
                
    # ...for each location
    
    print('Created {} sequences from {} images'.format(len(all_sequences),len(image_info)))
    
# ...create_sequences()



"""

 auckland_doc_to_json.py

 Convert Auckland DOC data set to COCO camera traps format.  This was
 for a training data set where class names were encoded in path names.

"""

#%% Constants and imports

import json
import os
import uuid
import datetime
from tqdm import tqdm

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db
from md_utils.path_utils import find_images, split_path, insert_before_extension

# Filenames will be stored in the output .json relative to this base dir
input_base_dir = 'y:\\'
output_base_dir = r'f:\auckland-doc'
output_json_filename = os.path.join(output_base_dir, 'auckland-doc-Maukahuka_Auckland_Island.json')

assert os.path.isdir(input_base_dir)
os.makedirs(output_base_dir,exist_ok=True)

output_encoding = 'utf-8'
read_image_sizes = True

info = {}
info['year'] = 2019
info['version'] = '1.0'
info['description'] = 'Auckaland DOC Camera Traps'
info['contributor'] = 'Auckland DOC'
info['date_created'] = str(datetime.date.today())


#%% Enumerate files

print('Enumerating files from {}'.format(input_base_dir))
image_files = find_images(input_base_dir, bRecursive=True)
print('Enumerated {} images'.format(len(image_files)))


#%% Assemble dictionaries

images = []
image_id_to_image = {}
annotations = []
categories = []

category_name_to_category = {}
category_id_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_id_to_category[0] = empty_category
categories.append(empty_category)
next_id = 1

behaviors = set()

# fn = image_files[0]; print(fn)
for fn in tqdm(image_files):

    # Typically y:\Maukahuka_Auckland_Island\1_Training\Winter_Trial_2019\cat\cat\eat\20190903_IDdY_34_E3_tmp_201908240051.JPG
    relative_path = os.path.relpath(fn,input_base_dir)
    tokens = split_path(fn)
    assert tokens[1] == 'Maukahuka_Auckland_Island'
    
    trainval_split = tokens[2]
    assert trainval_split in ['1_Training','2_Testing']
    
    # This data set has two top-level folders, "1_Training" (which has class names encoded
    # in paths) and "2_Testing" (which has no class information).
    if trainval_split == '2_Testing':
        category_name = 'test'
    else:
        category_name = tokens[-3]
        if category_name.startswith('2_'):
            category_name = category_name.replace('2_', '')
        category_name = category_name.lower().strip()

    if category_name not in category_name_to_category:        

        category_id = next_id
        next_id += 1
        category = {}
        category['id'] = category_id
        category['name'] = category_name
        category['count'] = 0
        categories.append(category)
        category_name_to_category[category_name] = category
        category_id_to_category[category_id] = category

    else:
        
        category = category_name_to_category[category_name]
        
    category_id = category['id']
    
    category['count'] += 1
    behavior = None
    if (category_name) != 'test':
        behavior = fn.split('\\')[-2]
        behaviors.add(behavior)            

    im = {}
    im['id'] = str(uuid.uuid1())
    im['file_name'] = relative_path
    image_id_to_image[im['id']] = im
    
    images.append(im)
    
    ann = {}
    
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']
    ann['category_id'] = category_id
    if behavior is not None:
        ann['behavior'] = behavior
    annotations.append(ann)

# ...for each image
    

#%% Write output .json

data = {}
data['info'] = info
data['images'] = images
data['annotations'] = annotations
data['categories'] = categories

json.dump(data, open(output_json_filename, 'w'), indent=2)
print('Finished writing json to {}'.format(output_json_filename))


#%% Write train/test .jsons

train_images = []; test_images = []
train_annotations = []; test_annotations = []

for ann in tqdm(annotations):
    category_id = ann['category_id']
    image_id = ann['image_id']
    category_name = category_id_to_category[category_id]['name']
    im = image_id_to_image[image_id]
    if category_name == 'test':
        test_images.append(im)
        test_annotations.append(ann)
    else:
        train_images.append(im)
        train_annotations.append(ann)

train_fn = insert_before_extension(output_json_filename,'train')
test_fn = insert_before_extension(output_json_filename,'test')

data['images'] = train_images
data['annotations'] = train_annotations
json.dump(data, open(train_fn, 'w'), indent=2)

data['images'] = test_images
data['annotations'] = test_annotations
json.dump(data, open(test_fn, 'w'), indent=2)

    
#%% Validate .json files

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = input_base_dir
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = True

sorted_categories, data, _ = integrity_check_json_db.integrity_check_json_db(output_json_filename, options)
sorted_categories, data, _ = integrity_check_json_db.integrity_check_json_db(train_fn, options)
sorted_categories, data, _ = integrity_check_json_db.integrity_check_json_db(test_fn, options)


#%% Preview labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 2000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['test']
html_output_file, image_db = visualize_db.visualize_db(db_path=output_json_filename,
                                                         output_dir=os.path.join(
                                                         output_base_dir, 'preview'),
                                                         image_base_dir=input_base_dir,
                                                         options=viz_options)
os.startfile(html_output_file)



"""

manage_video_batch.py

Notebook-esque script to manage the process of running a local batch of videos
through MD.  Defers most of the heavy lifting to manage_local_batch.py .

"""

#%% Imports and constants

import os

from md_utils import path_utils
from detection import video_utils

input_folder = '/datadrive/data'
output_folder_base = '/datadrive/frames'

assert os.path.isdir(input_folder)
os.makedirs(output_folder_base,exist_ok=True)


#%% Split videos into frames

assert os.path.isdir(input_folder)
os.makedirs(output_folder_base,exist_ok=True)

recursive = True
overwrite = True
n_threads = 4
every_n_frames = 10

frame_filenames_by_video,fs_by_video,video_filenames = \
    video_utils.video_folder_to_frames(input_folder=input_folder,
                                                              output_folder_base=output_folder_base,
                                                              recursive=recursive,
                                                              overwrite=overwrite,
                                                              n_threads=n_threads,
                                                              every_n_frames=every_n_frames,
                                                              parallelization_uses_threads=False)


#%% List frame files, break into folders

from collections import defaultdict

frame_files = path_utils.find_images(output_folder_base,True)
frame_files = [s.replace('\\','/') for s in frame_files]
print('Enumerated {} total frames'.format(len(frame_files)))

# Find unique (relative) folders
folder_to_frame_files = defaultdict(list)

# fn = frame_files[0]
for fn in frame_files:
    folder_name = os.path.dirname(fn)
    folder_name = os.path.relpath(folder_name,output_folder_base)
    folder_to_frame_files[folder_name].append(fn)

print('Found {} folders for {} files'.format(len(folder_to_frame_files),len(frame_files)))


#%% List videos

video_filenames = video_utils.find_videos(input_folder,recursive=True)
video_filenames = [os.path.relpath(fn,input_folder) for fn in video_filenames]
print('Input folder contains {} videos'.format(len(video_filenames)))


#%% Check for videos that are missing entirely

# list(folder_to_frame_files.keys())[0]
# video_filenames[0]

missing_videos = []

# fn = video_filenames[0]
for relative_fn in video_filenames:
    if relative_fn not in folder_to_frame_files:
        missing_videos.append(relative_fn)
        
print('{} of {} folders are missing frames entirely'.format(len(missing_videos),
                                                            len(video_filenames)))


#%% Check for videos with very few frames

min_frames_for_valid_video = 10

low_frame_videos = []

for folder_name in folder_to_frame_files.keys():
    frame_files = folder_to_frame_files[folder_name]
    if len(frame_files) < min_frames_for_valid_video:
        low_frame_videos.append(folder_name)

print('{} of {} folders have fewer than {} frames'.format(
    len(low_frame_videos),len(video_filenames),min_frames_for_valid_video))


#%% Print the list of videos that are problematic

print('Videos that could not be decoded:\n')

for fn in missing_videos:
    print(fn)
    
print('\nVideos with fewer than {} decoded frames:\n'.format(min_frames_for_valid_video))

for fn in low_frame_videos:
    print(fn)
    
    
#%% Process images like we would for any other camera trap job

# ...typically using manage_local_batch.py or manage_local_batch.ipynb, but do this however 
# you like, as long as you get a results file at the end.
#
# If you do RDE, remember to use the second folder from the bottom, rather than the
# bottom-most folder.


#%% Convert frame results to video results
    
from detection.video_utils import frame_results_to_video_results

filtered_output_filename = '/results/organization/stuff.json'
video_output_filename = filtered_output_filename.replace('.json','_aggregated.json')
frame_results_to_video_results(filtered_output_filename,video_output_filename)
    

#%% Confirm that the videos in the .json file are what we expect them to be

import json

with open(video_output_filename,'r') as f:
    video_results = json.load(f)

video_filenames_set = set(video_filenames)

filenames_in_video_results_set = set([im['file'] for im in video_results['images']])

for fn in filenames_in_video_results_set:
    assert fn in video_filenames_set
    

#%% Scrap

if False:
    
    pass

    #%% Render one or more sample videos...
        
    # ...while we still have the frames and detections around
    
    ## Imports
    
    from md_visualization import visualize_detector_output
    from detection.video_utils import frames_to_video    
    
    
    ## Constants and paths
    
    confidence_threshold = 0.2
    input_fs = 30
    
    filtered_output_filename = '/a/b/c/blah_detections.filtered_rde_0.150_0.850_10_1.000.json'
    video_fn_relative = '4.10cam6/IMG_0022.MP4'
    output_video_base = os.path.expanduser('~/tmp/video_preview')
    
    
    ## Filename handling
    
    video_fn_relative = video_fn_relative.replace('\\','/')
    video_fn_flat = video_fn_relative.replace('/','#')
    video_name = os.path.splitext(video_fn_flat)[0]
    output_video = os.path.join(output_video_base,'{}_detections.mp4'.format(video_name))        
    output_fs = input_fs / every_n_frames
    
    rendered_detections_folder = os.path.join(output_video_base,'rendered_detections_{}'.format(video_name))
    os.makedirs(rendered_detections_folder,exist_ok=True)
    
    
    ## Find frames corresponding to this video
    
    with open(filtered_output_filename,'r') as f:
        frame_results = json.load(f)
    
    frame_results_this_video = []
    
    # im = frame_results['images'][0]
    for im in frame_results['images']:
        if im['file'].replace('\\','/').startswith(video_fn_relative):
            frame_results_this_video.append(im)

    assert len(frame_results_this_video) > 0, \
        'No frame results matched {}'.format(video_fn_relative)
    print('Found {} matching frame results'.format(len(frame_results_this_video)))
    
    frame_results['images'] = frame_results_this_video
    
    frames_json = os.path.join(rendered_detections_folder,video_fn_flat + '.json')
    
    with open(frames_json,'w') as f:
        json.dump(frame_results,f,indent=1)        
        
        
    ## Render detections on those frames    
    
    detected_frame_files = visualize_detector_output.visualize_detector_output(
        detector_output_path=frames_json,
        out_dir=rendered_detections_folder,
        images_dir=output_folder_base,
        confidence_threshold=confidence_threshold,
        preserve_path_structure=True,
        output_image_width=-1)
        
    
    ## Render the output video
    
    frames_to_video(detected_frame_files, output_fs, output_video, codec_spec='h264')
    
    # from md_utils.path_utils import open_file; open_file(output_video)


    #%% Test a possibly-broken video
    
    fn = '/datadrive/tmp/video.AVI'
        
    fs = video_utils.get_video_fs(fn)
    print(fs)
    
    tmpfolder = '/home/user/tmp/frametmp'
    os.makedirs(tmpfolder,exist_ok=True)
    
    video_utils.video_to_frames(fn, tmpfolder, verbose=True, every_n_frames=10)
    
    
    #%% List videos in a folder
    
    input_folder = '/datadrive/tmp/organization/data'
    video_filenames = video_utils.find_videos(input_folder,recursive=True)


#%% End notebook: turn this script into a notebook (how meta!)

import os
import nbformat as nbf

if os.name == 'nt':
    git_base = r'c:\git'
else:
    git_base = os.path.expanduer('~/git')

input_py_file = git_base + '/MegaDetector/api/batch_processing/data_preparation/manage_video_batch.py'
assert os.path.isfile(input_py_file)
output_ipynb_file = input_py_file.replace('.py','.ipynb')

nb_header = '# Managing a local MegaDetector video batch'

nb_header += '\n'

nb_header += \
"""
This notebook represents an interactive process for running MegaDetector on large batches of videos, including typical and optional postprocessing steps.

This notebook is auto-generated from manage_video_batch.py (a cell-delimited .py file that is used the same way, typically in Spyder or VS Code).

"""

with open(input_py_file,'r') as f:
    lines = f.readlines()

nb = nbf.v4.new_notebook()
nb['cells'].append(nbf.v4.new_markdown_cell(nb_header))

i_line = 0

# Exclude everything before the first cell
while(not lines[i_line].startswith('#%%')):
    i_line += 1

current_cell = []

def write_code_cell(c):
    
    first_non_empty_line = None
    last_non_empty_line = None
    
    for i_code_line,code_line in enumerate(c):
        if len(code_line.strip()) > 0:
            if first_non_empty_line is None:
                first_non_empty_line = i_code_line
            last_non_empty_line = i_code_line
            
    # Remove the first [first_non_empty_lines] from the list
    c = c[first_non_empty_line:]
    last_non_empty_line -= first_non_empty_line
    c = c[:last_non_empty_line+1]
    
    nb['cells'].append(nbf.v4.new_code_cell('\n'.join(c)))
        
while(True):    
            
    line = lines[i_line].rstrip()
    
    if 'end notebook' in line.lower():
        break
    
    if lines[i_line].startswith('#%% '):
        if len(current_cell) > 0:
            write_code_cell(current_cell)
            current_cell = []
        markdown_content = line.replace('#%%','##')
        nb['cells'].append(nbf.v4.new_markdown_cell(markdown_content))
    else:
        current_cell.append(line)

    i_line += 1

# Add the last cell
write_code_cell(current_cell)

nbf.write(nb,output_ipynb_file)




"""

get_lila_image_counts.py

Count the number of images and bounding boxes with each label in one or more LILA datasets.

This script doesn't write these counts out anywhere other than the console, it's just intended
as a template for doing operations like this on LILA data.  get_lila_annotation_counts.py writes 
information out to a .json file, but it counts *annotations*, not *images*, for each category.

"""

#%% Constants and imports

import json
import os

from collections import defaultdict

from data_management.lila.lila_common import read_lila_metadata, read_metadata_file_for_dataset

# If None, will use all datasets
datasets_of_interest = None

# We'll write images, metadata downloads, and temporary files here
lila_local_base = os.path.expanduser('~/lila')

metadata_dir = os.path.join(lila_local_base,'metadata')
os.makedirs(metadata_dir,exist_ok=True)


#%% Download and parse the metadata file

metadata_table = read_lila_metadata(metadata_dir)


#%% Download and extract metadata for the datasets we're interested in

if datasets_of_interest is None:
    datasets_of_interest = list(metadata_table.keys())

for ds_name in datasets_of_interest:    
    metadata_table[ds_name]['json_filename'] = read_metadata_file_for_dataset(ds_name=ds_name,
                                                                         metadata_dir=metadata_dir,
                                                                         metadata_table=metadata_table)
    
    
#%% Count categories

ds_name_to_category_counts = {}

# ds_name = datasets_of_interest[0]
for ds_name in datasets_of_interest:
    
    category_to_image_count = {}
    category_to_bbox_count = {}
    
    print('Counting categories in: ' + ds_name)
    
    json_filename = metadata_table[ds_name]['json_filename']
    with open(json_filename, 'r') as f:
        data = json.load(f)
    
    categories = data['categories']
    category_ids = [c['id'] for c in categories]
    for c in categories:
        category_id_to_name = {c['id']:c['name'] for c in categories}
    annotations = data['annotations']
    images = data['images']
    
    for category_id in category_ids:        
        category_name = category_id_to_name[category_id]        
        category_to_image_count[category_name] = 0
        category_to_bbox_count[category_name] = 0
        
    image_id_to_category_names = defaultdict(set)
    
    # Go through annotations, marking each image with the categories that are present
    #
    # ann = annotations[0]
    for ann in annotations:
        
        category_name = category_id_to_name[ann['category_id']]
        image_id_to_category_names[ann['image_id']].add(category_name)

    # Now go through images and count categories
    category_to_count = defaultdict(int)
    
    # im = images[0]
    for im in images:
        categories_this_image = image_id_to_category_names[im['id']]
        for category_name in categories_this_image:
            category_to_count[category_name] += 1

    ds_name_to_category_counts[ds_name] = category_to_count
    
# ...for each dataset
    

#%% Print the results

for ds_name in ds_name_to_category_counts:
    
    print('\n** Category counts for {} **\n'.format(ds_name))
    
    category_to_count = ds_name_to_category_counts[ds_name]
    category_to_count = {k: v for k, v in sorted(category_to_count.items(), reverse=True, 
                                                 key=lambda item: item[1])}
    
    for category_name in category_to_count.keys():        
        print('{}: {}'.format(category_name,category_to_count[category_name]))



"""

 cacophony-thermal-importer.py

 Create data and metadata for LILA from the Cacophony thermal dataset.  Takes a folder
 of HDF files, and produces .json metadata, along with compressed/normalized videos for
 each HDF file.

 Source format notes for this dataset:
    
 https://docs.google.com/document/d/12sw5JtwdMf9MiXuNCBcvhvZ04Jwa1TH2Lf6LnJmF8Bk/edit

"""

#%% Imports and constants

import os
import h5py
import numpy as np
import json

from tqdm import tqdm
from copy import deepcopy
from collections import defaultdict
from multiprocessing.pool import Pool

import zipfile
from zipfile import ZipFile

import cv2

from md_utils.ct_utils import truncate_float
from md_utils import path_utils

base_dir = '/bigdata/home/sftp/cacophony-ferraro_/data/cacophony-thermal/'
output_base = os.path.expanduser('~/tmp/new-zealand-wildlife-thermal-imaging')
video_output_folder = os.path.join(output_base,'videos')
individual_metadata_output_folder = os.path.join(output_base,'individual-metadata')

os.makedirs(video_output_folder,exist_ok=True)
os.makedirs(individual_metadata_output_folder,exist_ok=True)

main_metadata_filename = 'new-zealand-wildlife-thermal-imaging.json'

# Every HDF file specifies a crop rectangle within which the pixels are trustworthy;
# in practice this is the same across all files.
expected_crop_rectangle = [1,1,159,119]

# Required attributes for each video
expected_clip_attributes = ['clip_id', 'crop_rectangle', 'ffc_frames', 'frame_temp_max', 
                            'frame_temp_mean', 'frame_temp_median', 'frame_temp_min', 'max_temp', 
                            'mean_temp', 'min_temp', 'num_frames', 'res_x', 'res_y', 'start_time',
                            'station_id']

# Attributes that may or may not be present for each video
optional_clip_attributes = ['temp_thresh','model']

# Required attributes for each trck
expected_track_attributes = ['end_frame', 'id', 'start_frame']

# Attributes that may or may not be present for each track
optional_track_attributes = ['human_tag', 'human_tag_confidence', 'human_tags', 
                             'human_tags_confidence', 'ai_tag', 'ai_tag_confidence']

labels_to_ignore_when_another_label_is_present = ['false-positive','unidentified','part','poor tracking']

frame_rate = 9

use_default_filtering = False
write_as_color = False

# codec = 'ffv1'
# codec = 'hfyu'
codec = 'h264'
overwrite_video = True

codec_to_extension = {'mp4v':'.mp4','ffv1':'.avi','hfyu':'.avi','h264':'.mp4'}

# Set to >0 to process only a subset of clips
debug_n = -1
n_workers = 16
confidence_digits = 3

# Standardize a few tag names
tag_mappings = {
    'bird/kiwi':'bird',
    'allbirds':'bird',
    'not identifiable':'unidentified',
    'part':'unidentified',
    'pest':'unidentified'
}    

# Discard tracks and labels that are below this confidence threshold.
confidence_threshold = 0.001


#%% Support functions

def remove_tracking_points(clip_metadata):
    """
    As a debugging convenience, take the metadata for a clip (after conversion to
    the output format) and remove the only field that makes it hard to read in a
    console (the track coordinates).
    """
    
    slim_metadata = deepcopy(clip_metadata)
    if 'tracks' in slim_metadata:
        for t in slim_metadata['tracks']:
            del t['points']
    return slim_metadata

    
def norm_image(image,vmin=None,vmax=None,do_normalization=True,stack_channels=True):
    """
    Normalize an MxN 2D numpy ndarray (may be any type, but typically uint16) into the range 
    0,255.  
    
    If stack_channels==True, return as an MxNx3 uint8 matrix (content is replicated across 
    all three channels).
    """
    
    if vmin is not None:
        assert vmax is not None
        assert vmax > vmin
    if vmax is not None:
        assert vmin is not None

    assert isinstance(image,np.ndarray)
    assert isinstance(image[0][0],np.uint16) or isinstance(image[0][0],np.float32), \
        'First pixel is of type {}'.format(type(image[0][0]))
    assert len(image.shape) == 2
    
    norm = np.float32(image)
    
    if do_normalization:
        
        if vmin is None:
            vmin = np.amin(image)
            vmax = np.amax(image)
            
        norm = 255 * (norm - vmin) / (vmax - vmin)

    norm = np.uint8(norm)
    norm = norm[:, :, np.newaxis]
    if stack_channels:
        norm = np.repeat(norm, 3, axis=2)
    return norm


#%% Enumerate files

all_files = path_utils.recursive_file_list(base_dir)
all_hdf_files_relative = [os.path.relpath(fn,base_dir) for fn in all_files if fn.lower().endswith('.hdf5')]

print('Found {} HDF files (of {} total files)'.format(
    len(all_hdf_files_relative),len(all_files)))


#%% Process one file

def process_file(fn_relative,verbose=False):
    """
    Read the HDF file, convert to video files with/without filtering, and return 
    a metadata dict for this file.
    """
    
    fn_abs = os.path.join(base_dir,fn_relative)
    
    clip_id = int(os.path.basename(fn_relative).split('.')[0])
    metadata_fn = os.path.join(individual_metadata_output_folder,str(clip_id) + '_metadata.json')
    
    clip_metadata = {}
    clip_metadata['hdf_filename'] = os.path.basename(fn_relative)    
    clip_metadata['id'] = clip_id
    clip_metadata['error'] = None
        
    try:
        h5f = h5py.File(fn_abs, 'r')
    except Exception as e:
        print('Could not open file {}: {}'.format(
            fn_relative,str(e)))
        clip_metadata['error'] = str(e)
        with open(metadata_fn,'w') as f:
            json.dump(clip_metadata,f,indent=1)        
        return clip_metadata    
    
    clip_attrs = h5f.attrs
    
    for s in expected_clip_attributes:
        assert s in clip_attrs
                
    assert clip_id == int(clip_attrs.get('clip_id'))
    assert os.path.basename(fn_relative).startswith(str(clip_id))
    
    station_id = clip_attrs.get('station_id')
    assert isinstance(station_id,np.int64)
    station_id = int(station_id)
    
    crop_rectangle = clip_attrs.get('crop_rectangle')
    assert len(crop_rectangle) == 4
    for i_coord in range(0,4):
        assert crop_rectangle[i_coord] == expected_crop_rectangle[i_coord]
      
    frames = h5f['frames']
    assert 'thermals' in frames
    
    # This is an HDF dataset of size n_frames,y,x
    thermal_frames = frames['thermals']
    assert len(thermal_frames.shape) == 3
    
    # If present, this is an HDF dataset of size y,x
    if 'background' in frames:
        background_frame = frames['background']
        assert len(background_frame.shape) == 2
        assert background_frame.shape[0] == thermal_frames.shape[1]
        assert background_frame.shape[1] == thermal_frames.shape[2]
    else:
        background_frame = None
    calibration_frame_indices = clip_attrs.get('ffc_frames')
    
    if len(calibration_frame_indices) > 0:
        assert max(calibration_frame_indices) < thermal_frames.shape[0]
    
    assert clip_attrs.get('num_frames') == thermal_frames.shape[0]
    assert clip_attrs.get('res_x') == thermal_frames.shape[2]
    assert clip_attrs.get('res_y') == thermal_frames.shape[1]    
    assert clip_attrs.get('model') in [None,'lepton3.5','lepton3']
        
    tracks = h5f['tracks']
    
    track_ids = list(tracks.keys())
    
    # List of dicts
    tracks_this_clip = []
    
    # i_track = 0; track_id = track_ids[i_track]
    for i_track,track_id in enumerate(track_ids):
        
        track = tracks[track_id]
        
        if 'human_tags' not in track.attrs.keys():
            continue
        
        track_info = {}
            
        # 'human_tags' is all the tags that were assigned to this track by humans
        # 'human_tags_confidence' is the confidence for each of those assignments
        #
        # If there is a clear "winner", 'human_tag' and 'human_tag' confidence will
        # identify the clear winner.
        if 'human_tag' in track.attrs.keys():
            
            assert 'human_tags' in track.attrs.keys()
            assert 'human_tags_confidence' in track.attrs.keys()
            assert 'human_tag_confidence' in track.attrs.keys()            
        
        track_tags = []
        
        if 'human_tags' in track.attrs.keys():
            
            assert 'human_tags_confidence' in track.attrs.keys()            
            assert len(track.attrs.get('human_tags_confidence')) == \
                   len(track.attrs.get('human_tags'))            
            
            human_tags_this_clip = list(track.attrs.get('human_tags'))
            human_tag_confidences_this_clip = list(track.attrs.get('human_tags_confidence'))
            
            for i_tag,tag in enumerate(human_tags_this_clip):
                assert isinstance(tag,str)
                tag_info = {}
                tag_info['label'] = tag
                conf = float(human_tag_confidences_this_clip[i_tag])
                tag_info['confidence'] = truncate_float(conf,confidence_digits)
                track_tags.append(tag_info)
                
        track_start_frame = int(round(track.attrs.get('start_frame')))
        track_end_frame = int(round(track.attrs.get('end_frame')))
        track_info['start_frame'] = track_start_frame
        track_info['end_frame'] = track_end_frame
        track_info['tags'] = track_tags
        
        # A list of x/y/frame tuples
        track_info['points'] = []
        
        for s in expected_track_attributes:
            assert s in track.attrs
        
        positions = track['regions']
        
        # Positions is an N x 7 matrix in which each row looks like:
        #
        # [left,top,right,bottom,frame_number,mass,blank_frame]
        #
        # The origin appears to be in the upper-left.
        # 
        # "blank_frame" indicates that the tracked object is not visible in this frame,
        # but was predicted from previous frames.
        assert positions.shape[1] == 7        
        
        # The number of items in the positions array should be equal to the length of the track, but this
        # can be off by a little when 'start_frame' and/or 'end_frame' are not integers.  Make sure this
        # is approximately true.
        
        # assert positions.shape[0] == 1 + (track.attrs.get('end_frame') - track.attrs.get('start_frame'))
        track_length_error = abs(positions.shape[0] - 
            (1 + (track.attrs.get('end_frame') - track.attrs.get('start_frame'))))
        assert track_length_error < 2
        
        # i_position = 0; position = positions[i_position]
        for i_position,position in enumerate(positions):
            
            left = float(position[0])
            top = float(position[1])
            right = float(position[2])
            bottom = float(position[3])
            frame_number = int(position[4])
            
            # TODO: I'm being lazy about the fact that these don't reflect the
            # pixels cropped out of the border.  IMO this is OK because for this dataset,
            # this is just an approximate set of coordinates used to disambiguate simultaneous 
            # areas of movement when multiple different labels are present in the same video.
            position_info = [left+float((right-left)/2),
                             top+float((bottom-top)/2),
                             int(frame_number)]
            track_info['points'].append(position_info)
            
            # In a small number of tracks, boxes are turned upside-down or left-over-right, 
            # we don't bother checking for coordinate validity in those tracks.
            if left <= right:
                assert left >= 0 and left < clip_attrs.get('res_x')
                assert right >= 0 and right < clip_attrs.get('res_x')
            
            if top <= bottom:
                assert top >= 0 and top < clip_attrs.get('res_y')
                assert bottom >= 0 and bottom < clip_attrs.get('res_y')
        
            # frame_number should be approximately equal to i_position + start_frame, but this
            # can be off by a little when 'start_frame' and/or 'end_frame' are not integers. 
            # Make sure this is approximately true.
            
            # assert frame_number == i_position + track.attrs.get('start_frame')
            frame_number_error = abs(frame_number - (i_position + track.attrs.get('start_frame')))
            assert frame_number_error <= 2
        
        # ...for each position in this track
        
        tracks_this_clip.append(track_info)
        
    # ...for each track ID                
    
    clip_metadata['tracks'] = tracks_this_clip
    
    assert len(human_tags_this_clip) > 0
    
    ffc_frames = clip_attrs.get('ffc_frames').tolist()
    if len(ffc_frames) > 0:
        assert max(ffc_frames) < thermal_frames.shape[0]
        n_ffc_frames = len(ffc_frames)
        n_frames = thermal_frames.shape[0]
        if verbose:
            if (n_ffc_frames / n_frames) > 0.2:
                print('Warning: in video {}, {} of {} frames are FFC frames (tags: {})'.format(
                    fn_relative,n_ffc_frames,n_frames,str(human_tags_this_clip)))                
    
    frames = h5f["frames"]
    
    if "background" in frames:
        background = frames["background"]
        background_frame_present = True        
    else:
        background = frames["thermals"][0]
        background_frame_present = False
    
    crop_rectangle = clip_attrs["crop_rectangle"]
    background = background[
        crop_rectangle[1]:crop_rectangle[3],
        crop_rectangle[0]:crop_rectangle[2]
    ]
    
    # Compute the median frame value
    #
    # (...which we may use for filtering)
    
    frames_array = np.array(frames['thermals'])
    frames_array = frames_array[:,crop_rectangle[1] : crop_rectangle[3], crop_rectangle[0] : crop_rectangle[2]]
    median_values = np.float32(np.median(frames_array,0))
    
    if (background_frame_present or use_default_filtering):
        background_for_filtering = background
    else:
        if verbose:
            print('No background present: using median values for background')
        background_for_filtering = median_values
    
    # Find the largest value by which any pixel in this video exceeds the background 
    #
    # (...which we may use for normalization)
    
    max_pixel_diff = 0
    
    for frame in frames["thermals"]:
        cropped_frame = frame[
            crop_rectangle[1]:crop_rectangle[3],
            crop_rectangle[0]:crop_rectangle[2]
        ]
        
        filtered_frame = np.float32(cropped_frame) - background_for_filtering
        max_pixel_diff_this_frame = np.amax(filtered_frame)
        if max_pixel_diff_this_frame > max_pixel_diff:
            max_pixel_diff = max_pixel_diff_this_frame
    
    filtered_frames = []
    original_frames = []
    
    # i_frame = 0; frame = frames["thermals"][i_frame]
    for i_frame,frame in enumerate(frames["thermals"]):
        
        cropped_frame = frame[crop_rectangle[1] : crop_rectangle[3], crop_rectangle[0] : crop_rectangle[2]]

        # Subtract the background frame
        filtered_frame = np.float32(cropped_frame) - background_for_filtering
        
        # Assume that nothing can be cooler than the background
        filtered_frame[filtered_frame < 0] = 0
        
        # Normalize filtered frame (and convert to three channels)
        
        if use_default_filtering:
            filtered_frame = norm_image(filtered_frame,stack_channels=write_as_color)
        else:        
            filtered_frame = norm_image(filtered_frame,vmin=0,vmax=max_pixel_diff,stack_channels=write_as_color)
        
        # Normalize original frame (and convert to three channels)
        
        original_frame = norm_image(cropped_frame,stack_channels=write_as_color)
                
        filtered_frames.append(filtered_frame)        
        original_frames.append(original_frame)
        
    # ...for each frame

    # filtered_frames[0].shape[1] is 158, clip_attrs.get('res_x') is 160, ergo shape is h,w
    video_w = filtered_frames[0].shape[1]
    video_h = filtered_frames[0].shape[0]
    
    clip_metadata['width'] = video_w
    clip_metadata['height'] = video_h
    clip_metadata['frame_rate'] = frame_rate
    
    filtered_video_fn = os.path.join(video_output_folder,str(clip_id) + '_filtered' + codec_to_extension[codec])        
    unfiltered_video_fn = os.path.join(video_output_folder,str(clip_id) + codec_to_extension[codec])    
    
    if overwrite_video or (not os.path.isfile(filtered_video_fn)):
        
        filtered_video_out = cv2.VideoWriter(filtered_video_fn, cv2.VideoWriter_fourcc(*codec), frame_rate, 
                              (video_w, video_h), isColor=write_as_color)
        
        for i_frame,filtered_frame in enumerate(filtered_frames): 
            filtered_video_out.write(filtered_frame)
        filtered_video_out.release()
    
    if overwrite_video or (not os.path.isfile(unfiltered_video_fn)):
        
        unfiltered_video_out = cv2.VideoWriter(unfiltered_video_fn, cv2.VideoWriter_fourcc(*codec), frame_rate, 
                              (video_w, video_h), isColor=write_as_color)
            
        for i_frame,frame in enumerate(original_frames): 
            unfiltered_video_out.write(frame)
        unfiltered_video_out.release()
        
    labels_this_clip = set()
    
    
    ## Do some cleanup of tracks and track labels
    
    valid_tracks = []
    
    for track_info in clip_metadata['tracks']:
    
        valid_tags = []
        
        # Replace some tags with standardized names (e.g. map "allbirds" to "bird")
        for tag in track_info['tags']:
            if tag['label'] in tag_mappings:
                tag['label'] = tag_mappings[tag['label']]
                
            # Discard tags below the minimum confidence
            if tag['confidence'] >= confidence_threshold:
                valid_tags.append(tag)
            else:
                # TODO
                print('Zero-confidence tag in {}'.format(fn_relative))                
                
        track_info['tags'] = valid_tags
        
        # Don't keep any tracks that had no tags above the minimum confidence
        if len(valid_tags) > 0:
            valid_tracks.append(track_info)
        else:
            # TODO
            print('Invalid track in {}'.format(fn_relative))
        
    # ...for each track
    
    if (len(clip_metadata['tracks']) > 0) and (len(valid_tracks) == 0):
        # TODO
        print('Removed all tracks from {}'.format(fn_relative))
        
    clip_metadata['tracks'] = valid_tracks
                
    # Build up the list of labels for this clip
    for track_info in clip_metadata['tracks']:
        for tag in track_info['tags']:
            tag_label = tag['label']            
            labels_this_clip.add(tag_label)                

    clip_metadata['labels'] = sorted(list(labels_this_clip))
    
    metadata_fn = os.path.join(individual_metadata_output_folder,str(clip_id) + '_metadata.json')
    
    # clip_metadata['id'] = clip_id
    # clip_metadata['hdf_filename'] = os.path.basename(fn_relative)
    
    clip_metadata['video_filename'] = os.path.basename(unfiltered_video_fn)
    clip_metadata['filtered_video_filename'] = os.path.basename(filtered_video_fn)
    clip_metadata['location'] = station_id
    clip_metadata['calibration_frames'] = ffc_frames
    clip_metadata['metadata_filename'] = os.path.basename(metadata_fn)
    
    with open(metadata_fn,'w') as f:
        json.dump(clip_metadata,f,indent=1)
            
    return clip_metadata

# ...process_file(...)


#%% Process files

n_workers = 16

if debug_n > 0:
    files_to_process = all_hdf_files_relative[0:debug_n]
else:
    files_to_process = all_hdf_files_relative
    
if n_workers <= 1:
    
    all_clip_metadata = []    
    for i_file,fn_relative in tqdm(enumerate(files_to_process),total=len(files_to_process)):    
        clip_metadata = process_file(fn_relative)
        all_clip_metadata.append(clip_metadata)
        
else:
    
    pool = Pool(n_workers)
    all_clip_metadata = list(tqdm(pool.imap(process_file,files_to_process),
                                  total=len(files_to_process)))
        
    
#%% Postprocessing

failed_file_to_error = {}

label_to_video_count = defaultdict(int)

# clip_metadata = all_clip_metadata[0]
for clip_metadata in all_clip_metadata:
    
    if clip_metadata['error'] is not None:
        failed_file_to_error[clip_metadata['hdf_filename']] = clip_metadata['error']
        continue

    labels_this_clip = set()
    
    # track_info = clip_metadata['tracks'][0]
    for track_info in clip_metadata['tracks']:
        for tag in track_info['tags']:
            tag_label = tag['label']
            labels_this_clip.add(tag_label)
    
    for label in labels_this_clip:
        label_to_video_count[label] += 1
                    
# ...for each clip

label_to_video_count = {k: v for k, v in sorted(label_to_video_count.items(), 
                                                key=lambda item: item[1], reverse=True)}

print('Failed to open {} of {} files'.format(
    len(failed_file_to_error),len(all_hdf_files_relative)))

print('Labels:\n')

for label in label_to_video_count:
    print('{}: {}'.format(label,label_to_video_count[label]))


#%% Write count .csv

count_csv_file_name = os.path.join(output_base,'new-zealand-wildlife-thermal-imaging-counts.csv')

with open(count_csv_file_name,'w') as f:
    f.write('label,count\n')
    for label in label_to_video_count:
        f.write('{},{}\n'.format(label,label_to_video_count[label]))        


#%% Build and zip the main .json file

main_metadata_filename_abs = os.path.join(output_base,main_metadata_filename)

info = {}
info['version'] = '1.0.0'
info['description'] = 'New Zealand Thermal Wildlife Imaging'
info['contributor'] = 'Cacophony Project'

main_metadata = {}
main_metadata['info'] = info
main_metadata['clips'] = []

# clip_metadata = all_clip_metadata[0]
for clip_metadata in tqdm(all_clip_metadata):
    slim_metadata = remove_tracking_points(clip_metadata)
    
    if 'tracks' in slim_metadata:
        for track in slim_metadata['tracks']:
            for tag in track['tags']:
                tag['confidence'] = truncate_float(tag['confidence'],confidence_digits)
                
    main_metadata['clips'].append(slim_metadata)
    
with open(main_metadata_filename_abs,'w') as f:
    json.dump(main_metadata,f,indent=1)

zip_file_name = main_metadata_filename_abs.replace('.json','-metadata.json.zip')
with ZipFile(zip_file_name,'w',zipfile.ZIP_DEFLATED) as zipf:
    zipf.write(main_metadata_filename_abs,
               arcname=os.path.basename(main_metadata_filename_abs),
               compresslevel=9,compress_type=zipfile.ZIP_DEFLATED)


#%% Create a zipfile containing videos, main metadata, and individual metadata

zip_file_name = os.path.join(output_base,'new-zealand-wildlife-thermal-imaging.zip')

all_files = path_utils.recursive_file_list(output_base)
all_files_relative = [os.path.relpath(fn,output_base) for fn in all_files]
all_files_to_zip_relative = [fn for fn in all_files_relative if \
                             (\
                              ('individual-metadata/' in fn) or \
                              ('videos/' in fn) or \
                              (fn.endswith('.json'))
                             )]

print('Zipping {} files (of {} total files)'.format(len(all_files_to_zip_relative),len(all_files)))

with ZipFile(zip_file_name,'w',zipfile.ZIP_DEFLATED) as zipf:
    for fn_relative in tqdm(all_files_to_zip_relative):
        fn_abs = os.path.join(output_base,fn_relative)
        if fn_abs.endswith('.mp4'):
            zipf.write(fn_abs,arcname=fn_relative,compresslevel=0,compress_type=zipfile.ZIP_STORED)
        else:
            zipf.write(fn_abs,arcname=fn_relative,compresslevel=9,compress_type=zipfile.ZIP_DEFLATED)


#%% Scrap

if False:

    pass

    #%% Process one file

    # i_file = 110680; fn_relative = all_hdf_files_relative[i_file]        
    # i_file = 8; fn_relative = all_hdf_files_relative[i_file]    
    
    fn_relative = [fn for fn in all_hdf_files_relative if '450281' in fn][0]
    
    clip_metadata = process_file(fn_relative)
    
    
    #%% Move individual metadata files

    source_folder = base_dir
    target_folder = os.path.expanduser('~/tmp/cacophony-thermal-out-individual-metadata')
    assert os.path.isdir(source_folder) and os.path.isdir(target_folder)

    from md_utils import path_utils
    all_files = path_utils.recursive_file_list(source_folder)
    files_to_move = [fn for fn in all_files if '_metadata.json' in fn]
    print('Moving {} of {} files'.format(len(files_to_move),len(all_files)))

    import shutil
    # source_fn = files_to_move[0]
    for source_fn in tqdm(files_to_move):
        target_fn = os.path.join(target_folder,os.path.basename(source_fn))
        shutil.move(source_fn,target_fn)        


    #%% Choose a random video with a particular label

    target_label = 'pukeko'    
    target_clips = []
    
    for clip_metadata in all_clip_metadata:
        
        if clip_metadata['error'] is not None:
            continue
    
        labels_this_clip = set()
        
        # track_info = clip_metadata['tracks'][0]
        for track_info in clip_metadata['tracks']:
            for tag in track_info['tags']:
                tag_label = tag['label']
                labels_this_clip.add(tag_label)
            
        if target_label in labels_this_clip:            
            target_clips.append(clip_metadata)
    
    print('Found {} matches'.format(len(target_clips)))

    import random
    selected_clip = random.choice(target_clips)
    filtered_video_filename = selected_clip['filtered_video_filename']
    video_filename = selected_clip['video_filename']
    
    from md_utils.path_utils import open_file
    # open_file(os.path.join(output_base,video_filename))
    open_file(os.path.join(output_base,filtered_video_filename))
    
    # import clipboard; clipboard.copy(os.path.join(output_base,video_filename))


    #%% Look for clips with multiple different labels
    
    for i_clip,clip_metadata in enumerate(all_clip_metadata): 
        
        if clip_metadata['error'] is not None:
            continue
    
        labels_this_clip = set()
        
        # track_info = clip_metadata['tracks'][0]
        for track_info in clip_metadata['tracks']:
            for tag in track_info['tags']:
                tag_label = tag['label']
                if tag_label not in labels_to_ignore_when_another_label_is_present:
                    labels_this_clip.add(tag_label)                
    
        assert len(labels_this_clip) <= 3
                
        if len(labels_this_clip) > 1:
            print('Clip {} has {} labels: {}'.format(
                i_clip,len(labels_this_clip),str(labels_this_clip)))
        
        # remove_tracking_points(clip_metadata)
        
        
    #%% Add the .json filename to each clip in all_clip_metadata
    
    for i_clip,clip_metadata in tqdm(enumerate(all_clip_metadata),
                                     total=len(all_clip_metadata)): 
    
        clip_metadata['metadata_filename'] = clip_metadata['hdf_filename'].replace('.hdf5',
                                                                                     '_metadata.json')
        
        
    #%% Add a "labels" field to each .json file
    
    # This was only necessary during debugging; this is added in the main loop now.
    
    for i_clip,clip_metadata in tqdm(enumerate(all_clip_metadata),
                                     total=len(all_clip_metadata)): 
        
        if clip_metadata['error'] is not None:
            continue
    
        labels_this_clip = set()
        
        # track_info = clip_metadata['tracks'][0]
        for track_info in clip_metadata['tracks']:
            for tag in track_info['tags']:
                tag_label = tag['label']
                # if tag_label not in labels_to_ignore_when_another_label_is_present:
                if True:
                    labels_this_clip.add(tag_label)                
    
        clip_metadata['labels'] = sorted(list(labels_this_clip))
        
        json_filename = os.path.join(output_base,str(clip_metadata['id']) + '_metadata.json')
        assert os.path.isfile(json_filename)
                
        with open(json_filename,'w') as f:
            json.dump(clip_metadata,f,indent=1)



"""

 nacti_fieldname_adjustments.py

 NACTI metadata was posted with "filename" in images instead of "file_name", and
 used string (rather than int) category IDs (in categories, but not in annotations).

 This script fixes those issues and rev's the version number.

"""

#%% Constants and environment

import json
import os

inputJsonFile = r'/datadrive1/nacti_metadata_orig.json'
outputJsonFile = r'/datadrive1/nacti_metadata.json'

assert os.path.isfile(inputJsonFile)


#%% Read .json file

with open(inputJsonFile,'r') as f:
    data = json.load(f)

images = data['images']
annotations = data['annotations']
categories = data['categories']
info = data['info']

print('Finished reading input .json')


#%% Rev version number, update field names and types

assert(info['version'] == 1.0)
info['version'] = 1.1
nFilenameConversions = 0
nCatConversions = 0
nAnnConversions = 0

for image in images:
    
    assert 'path' in image and isinstance(image['path'],str)
    image['file_name'] = image['path']
    del image['path']
    nFilenameConversions += 1
    assert 'seq_no' in image
    del image['seq_no']
    assert 'width' in image and isinstance(image['width'],str)
    assert 'height' in image and isinstance(image['height'],str)
    image['width'] = int(image['width']) 
    image['height'] = int(image['height']) 
    
for cat in categories:
    
    assert 'id' in cat and isinstance(cat['id'],str)
    cat['id'] = int(cat['id'])
    nCatConversions += 1

for ann in annotations:
    
    assert 'id' in ann and isinstance(ann['id'],str)
    assert 'category_id' in ann and isinstance(ann['category_id'],str)
    ann['category_id'] = int(ann['category_id'])
    nAnnConversions += 1
    
print('Finished checking data, converted {} filename fields, {} category IDs, {} annotation category IDs'.format(
        nFilenameConversions,nCatConversions,nAnnConversions))


#%% Write json file
    
json.dump(data, open(outputJsonFile, 'w'), indent=4)

print('Finished writing output .json to {}'.format(outputJsonFile))




"""

visualize_db.py

Outputs an HTML page visualizing annotations (class labels and/or bounding boxes)
on a sample of images in a database in the COCO Camera Traps format.

"""

#%% Imports

import argparse
import inspect
import random
import json
import math
import os
import sys
import time

import pandas as pd
import numpy as np

import humanfriendly

from itertools import compress
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from tqdm import tqdm

from md_utils.write_html_image_list import write_html_image_list
from data_management.cct_json_utils import IndexedJsonDb

import md_visualization.visualization_utils as vis_utils


#%% Settings

class DbVizOptions:
    """
    Parameters controlling the behavior of visualize_db().
    """
    
    #: Number of images to sample from the database, or None to visualize all images
    num_to_visualize = None
    
    #: Target size for rendering; set either dimension to -1 to preserve aspect ratio.
    #:
    #: If viz_size is None or (-1,-1), the original image size is used.
    viz_size = (800, -1)
    
    #: HTML rendering options; see write_html_image_list for details
    #:
    #:The most relevant option one might want to set here is:
    #:
    #: htmlOptions['maxFiguresPerHtmlFile']
    #:
    #: ...which can be used to paginate previews to a number of images that will load well
    #: in a browser (5000 is a reasonable limit).
    htmlOptions = write_html_image_list()
    
    #: Whether to sort images by filename (True) or randomly (False)
    sort_by_filename = True
    
    #: Only show images that contain bounding boxes
    trim_to_images_with_bboxes = False
    
    #: Random seed to use for sampling images
    random_seed = 0
    
    #: Should we include Web search links for each category name?    
    add_search_links = False
    
    #: Should each thumbnail image link back to the original image?
    include_image_links = False
    
    #: Should there be a text link back to each original image?
    include_filename_links = False
    
    #: Line width in pixels
    box_thickness = 4
    
    #: Number of pixels to expand each bounding box
    box_expansion = 0
    
    #: Only include images that contain annotations with these class names (not IDs)
    #:
    #: Mutually exclusive with classes_to_exclude
    classes_to_include = None
    
    #: Exclude images that contain annotations with these class names (not IDs)
    #:
    #: Mutually exclusive with classes_to_include
    classes_to_exclude = None
    
    #: Special tag used to say "show me all images with multiple categories"
    #:
    #: :meta private:
    multiple_categories_tag = '*multiple*'

    #: We sometimes flatten image directories by replacing a path separator with 
    #: another character.  Leave blank for the typical case where this isn't necessary.
    pathsep_replacement = '' # '~'

    #: Parallelize rendering across multiple workers
    parallelize_rendering = False
    
    #: In theory, whether to parallelize with threads (True) or processes (False), but
    #: process-based parallelization in this function is currently unsupported
    parallelize_rendering_with_threads = True
    
    #: Number of workers to use for parallelization; ignored if parallelize_rendering
    #: is False
    parallelize_rendering_n_cores = 25
        
    #: Should we show absolute (True) or relative (False) paths for each image?
    show_full_paths = False
    
    #: Set to False to skip existing images
    force_rendering = True
    
    #: Enable additionald debug console output
    verbose = False
    

#%% Helper functions

def _image_filename_to_path(image_file_name, image_base_dir, pathsep_replacement=''):
    """
    Translates the file name in an image entry in the json database to a path, possibly doing
    some manipulation of path separators.
    """
    
    if len(pathsep_replacement) > 0:
        image_file_name = os.path.normpath(image_file_name).replace(os.pathsep,pathsep_replacement)        
    return os.path.join(image_base_dir, image_file_name)


#%% Core functions

def visualize_db(db_path, output_dir, image_base_dir, options=None):
    """
    Writes images and html to output_dir to visualize the annotations in a .json file.
    
    Args:
        db_path (str or dict): the .json filename to load, or a previously-loaded database
        image_base_dir (str): the folder where the images live; filenames in [db_path] should
            be relative to this folder.
        options (DbVizOptions, optional): See DbVizOptions for details
    
    Returns:
        tuple: A length-two tuple containing (the html filename) and (the loaded database).        
    """    
    
    if options is None:
        options = DbVizOptions()
    
    if not options.parallelize_rendering_with_threads:
        print('Warning: process-based parallelization is not yet supported by visualize_db')
        options.parallelize_rendering_with_threads = True
        
    if image_base_dir.startswith('http'):
        if not image_base_dir.endswith('/'):
            image_base_dir += '/'
    else:
        assert(os.path.isdir(image_base_dir))
            
    os.makedirs(os.path.join(output_dir, 'rendered_images'), exist_ok=True)
    
    if isinstance(db_path,str):
        assert(os.path.isfile(db_path))    
        print('Loading database from {}...'.format(db_path))
        image_db = json.load(open(db_path))
        print('...done')
    elif isinstance(db_path,dict):
        print('Using previously-loaded DB')
        image_db = db_path
    else:
        raise ValueError('Illegal dictionary or filename')    
        
    annotations = image_db['annotations']
    images = image_db['images']
    categories = image_db['categories']
    
    # Optionally remove all images without bounding boxes, *before* sampling
    if options.trim_to_images_with_bboxes:
        
        bHasBbox = [False] * len(annotations)
        for iAnn,ann in enumerate(annotations):
            if 'bbox' in ann:
                assert isinstance(ann['bbox'],list)
                bHasBbox[iAnn] = True
        annotationsWithBboxes = list(compress(annotations, bHasBbox))
        
        imageIDsWithBboxes = [x['image_id'] for x in annotationsWithBboxes]
        imageIDsWithBboxes = set(imageIDsWithBboxes)
        
        bImageHasBbox = [False] * len(images)
        for iImage,image in enumerate(images):
            imageID = image['id']
            if imageID in imageIDsWithBboxes:
                bImageHasBbox[iImage] = True
        imagesWithBboxes = list(compress(images, bImageHasBbox))
        images = imagesWithBboxes
                
    # Optionally include/remove images with specific labels, *before* sampling
        
    assert (not ((options.classes_to_exclude is not None) and \
                 (options.classes_to_include is not None))), \
                 'Cannot specify an inclusion and exclusion list'
        
    if (options.classes_to_exclude is not None) or (options.classes_to_include is not None):
     
        print('Indexing database')
        indexed_db = IndexedJsonDb(image_db)
        bValidClass = [True] * len(images)        
        for iImage,image in enumerate(images):
            classes = indexed_db.get_classes_for_image(image)
            if options.classes_to_exclude is not None:
                for excludedClass in options.classes_to_exclude:
                    if excludedClass in classes:
                       bValidClass[iImage] = False
                       break
            elif options.classes_to_include is not None:
                bValidClass[iImage] = False
                if options.multiple_categories_tag in options.classes_to_include:
                    if len(classes) > 1:
                        bValidClass[iImage] = True        
                if not bValidClass[iImage]:
                    for c in classes:
                        if c in options.classes_to_include:
                            bValidClass[iImage] = True
                            break                        
            else:
                raise ValueError('Illegal include/exclude combination')
                
        imagesWithValidClasses = list(compress(images, bValidClass))
        images = imagesWithValidClasses    
    
    # ...if we need to include/exclude categories
    
    # Put the annotations in a dataframe so we can select all annotations for a given image
    print('Creating data frames')
    df_anno = pd.DataFrame(annotations)
    df_img = pd.DataFrame(images)
    
    # Construct label map
    label_map = {}
    for cat in categories:
        label_map[int(cat['id'])] = cat['name']
    
    # Take a sample of images
    if options.num_to_visualize is not None:
        if options.num_to_visualize > len(df_img):
            print('Warning: asked to visualize {} images, but only {} are available, keeping them all'.\
                  format(options.num_to_visualize,len(df_img)))
        else:                  
            df_img = df_img.sample(n=options.num_to_visualize,random_state=options.random_seed)
    
    images_html = []
    
    # Set of dicts representing inputs to render_db_bounding_boxes:
    #
    # bboxes, boxClasses, image_path
    rendering_info = []
    
    print('Preparing rendering list')
    
    for iImage,img in tqdm(df_img.iterrows(),total=len(df_img)):
        
        img_id = img['id']
        assert img_id is not None
        
        img_relative_path = img['file_name']
        
        if image_base_dir.startswith('http'):
            img_path = image_base_dir + img_relative_path
        else:
            img_path = os.path.join(image_base_dir, 
                                    _image_filename_to_path(img_relative_path, image_base_dir))
    
        annos_i = df_anno.loc[df_anno['image_id'] == img_id, :] # all annotations on this image
    
        bboxes = []
        boxClasses = []
        
        # All the class labels we've seen for this image (with out without bboxes)
        imageCategories = set()
        
        annotationLevelForImage = ''
        
        # Iterate over annotations for this image
        # iAnn = 0; anno = annos_i.iloc[iAnn]
        for iAnn,anno in annos_i.iterrows():
        
            if 'sequence_level_annotation' in anno:
                bSequenceLevelAnnotation = anno['sequence_level_annotation']
                if bSequenceLevelAnnotation:
                    annLevel = 'sequence'
                else:
                    annLevel = 'image'
                if annotationLevelForImage == '':
                    annotationLevelForImage = annLevel
                elif annotationLevelForImage != annLevel:
                    annotationLevelForImage = 'mixed'
                    
            categoryID = anno['category_id']
            categoryName = label_map[categoryID]
            if options.add_search_links:
                categoryName = categoryName.replace('"','')
                categoryName = '<a href="https://www.google.com/search?tbm=isch&q={}">{}</a>'.format(
                    categoryName,categoryName)
            imageCategories.add(categoryName)
            
            if 'bbox' in anno:
                bbox = anno['bbox']        
                if isinstance(bbox,float):
                    assert math.isnan(bbox), "I shouldn't see a bbox that's neither a box nor NaN"
                    continue
                bboxes.append(bbox)
                boxClasses.append(anno['category_id'])
        
        # ...for each of this image's annotations
        
        imageClasses = ', '.join(imageCategories)
                
        img_id_string = str(img_id).lower()        
        file_name = '{}_gt.jpg'.format(os.path.splitext(img_id_string)[0])
        
        # Replace characters that muck up image links
        illegal_characters = ['/','\\',':','\t','#',' ','%']
        for c in illegal_characters:
            file_name = file_name.replace(c,'~')
        
        rendering_info.append({'bboxes':bboxes, 'boxClasses':boxClasses, 'img_path':img_path,
                               'output_file_name':file_name})
                
        labelLevelString = ' '
        if len(annotationLevelForImage) > 0:
            labelLevelString = ' (annotation level: {})'.format(annotationLevelForImage)
            
        if 'frame_num' in img and 'seq_num_frames' in img:
            frameString = ' frame: {} of {}, '.format(img['frame_num'],img['seq_num_frames'])
        elif 'frame_num' in img:
            frameString = ' frame: {}, '.format(img['frame_num'])
        else:
            frameString = ' '
        
        if options.show_full_paths:
            filename_text = img_path
        else:
            filename_text = img_relative_path
        if options.include_filename_links:
            filename_text = '<a href="{}">{}</a>'.format(img_path,filename_text)
            
        flagString = ''
        
        def isnan(x):
            return (isinstance(x,float) and np.isnan(x))
        
        if ('flags' in img) and (not isnan(img['flags'])):
            flagString = ', flags: {}'.format(str(img['flags']))
            
        # We're adding html for an image before we render it, so it's possible this image will
        # fail to render.  For applications where this script is being used to debua a database
        # (the common case?), this is useful behavior, for other applications, this is annoying.
        image_dict = \
        {
            'filename': '{}/{}'.format('rendered_images', file_name),
            'title': '{}<br/>{}, num boxes: {}, {}class labels: {}{}{}'.format(
                filename_text, img_id, len(bboxes), frameString, imageClasses, labelLevelString, flagString),
            'textStyle': 'font-family:verdana,arial,calibri;font-size:80%;' + \
                'text-align:left;margin-top:20;margin-bottom:5'
        }
        if options.include_image_links:
            image_dict['linkTarget'] = img_path
            
        images_html.append(image_dict)
    
    # ...for each image

    def render_image_info(rendering_info):
        
        img_path = rendering_info['img_path']
        bboxes = rendering_info['bboxes']
        bboxClasses = rendering_info['boxClasses']
        output_file_name = rendering_info['output_file_name']
        output_full_path = os.path.join(output_dir, 'rendered_images', output_file_name)
        
        if (os.path.isfile(output_full_path)) and (not options.force_rendering):
            if options.verbose:
                print('Skipping existing image {}'.format(output_full_path))
            return True
            
        if not img_path.startswith('http'):
            if not os.path.exists(img_path):
                print('Image {} cannot be found'.format(img_path))
                return False
            
        try:
            original_image = vis_utils.open_image(img_path)
            original_size = original_image.size
            if (options.viz_size is None) or (options.viz_size[0] == -1 and options.viz_size[1] == -1):
                image = original_image
            else:
                image = vis_utils.resize_image(original_image, options.viz_size[0],
                                               options.viz_size[1])
        except Exception as e:
            print('Image {} failed to open, error: {}'.format(img_path, e))
            return False
            
        vis_utils.render_db_bounding_boxes(boxes=bboxes, classes=bboxClasses,
                                           image=image, original_size=original_size,
                                           label_map=label_map,
                                           thickness=options.box_thickness,
                                           expansion=options.box_expansion)
        
        image.save(output_full_path)
                
        return True
    
    # ...def render_image_info
    
    print('Rendering images')
    start_time = time.time()
    
    if options.parallelize_rendering:
        
        if options.parallelize_rendering_with_threads:
            worker_string = 'threads'
        else:
            worker_string = 'processes'
            
        if options.parallelize_rendering_n_cores is None:
            if options.parallelize_rendering_with_threads:
                pool = ThreadPool()
            else:
                pool = Pool()
        else:
            if options.parallelize_rendering_with_threads:
                pool = ThreadPool(options.parallelize_rendering_n_cores)
            else:
                pool = Pool(options.parallelize_rendering_n_cores)
            print('Rendering images with {} {}'.format(options.parallelize_rendering_n_cores,
                                                       worker_string))            
        rendering_success = list(tqdm(pool.imap(render_image_info, rendering_info),
                                 total=len(rendering_info)))
        
    else:
        
        rendering_success = []
        for file_info in tqdm(rendering_info):        
            rendering_success.append(render_image_info(file_info))
            
    elapsed = time.time() - start_time
    
    print('Rendered {} images in {} ({} successful)'.format(
        len(rendering_info),humanfriendly.format_timespan(elapsed),sum(rendering_success)))
        
    if options.sort_by_filename:    
        images_html = sorted(images_html, key=lambda x: x['filename'])
    else:
        random.shuffle(images_html)
        
    htmlOutputFile = os.path.join(output_dir, 'index.html')
    
    htmlOptions = options.htmlOptions
    if isinstance(db_path,str):
        htmlOptions['headerHtml'] = '<h1>Sample annotations from {}</h1>'.format(db_path)
    else:
        htmlOptions['headerHtml'] = '<h1>Sample annotations</h1>'
        
    write_html_image_list(
            filename=htmlOutputFile,
            images=images_html,
            options=htmlOptions)

    print('Visualized {} images, wrote results to {}'.format(len(images_html),htmlOutputFile))
    
    return htmlOutputFile,image_db

# def visualize_db(...)
    
    
#%% Command-line driver
    
# Copy all fields from a Namespace (i.e., the output from parse_args) to an object.  
#
# Skips fields starting with _.  Does not check existence in the target object.
def args_to_object(args, obj):
    
    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)


def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('db_path', action='store', type=str, 
                        help='.json file to visualize')
    parser.add_argument('output_dir', action='store', type=str, 
                        help='Output directory for html and rendered images')
    parser.add_argument('image_base_dir', action='store', type=str, 
                        help='Base directory (or URL) for input images')

    parser.add_argument('--num_to_visualize', action='store', type=int, default=None, 
                        help='Number of images to visualize (randomly drawn) (defaults to all)')
    parser.add_argument('--random_sort', action='store_true', 
                        help='Sort randomly (rather than by filename) in output html')
    parser.add_argument('--trim_to_images_with_bboxes', action='store_true', 
                        help='Only include images with bounding boxes (defaults to false)')
    parser.add_argument('--random_seed', action='store', type=int, default=None,
                        help='Random seed for image selection')
    parser.add_argument('--pathsep_replacement', action='store', type=str, default='',
                        help='Replace path separators in relative filenames with another ' + \
                             'character (frequently ~)')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
            
    args = parser.parse_args()
    
    # Convert to an options object
    options = DbVizOptions()
    args_to_object(args, options)
    if options.random_sort:
        options.sort_by_filename = False
        
    visualize_db(options.db_path,options.output_dir,options.image_base_dir,options) 

if __name__ == '__main__':    
    main()


#%% Interactive driver

if False:
    
    #%%
    
    db_path = r'e:\wildlife_data\missouri_camera_traps\missouri_camera_traps_set1.json'
    output_dir = r'e:\wildlife_data\missouri_camera_traps\preview'
    image_base_dir = r'e:\wildlife_data\missouri_camera_traps'
    
    options = DbVizOptions()
    options.num_to_visualize = 100
    
    htmlOutputFile,db = visualize_db(db_path,output_dir,image_base_dir,options)
    # os.startfile(htmlOutputFile)



########
#
# copy_checkpoints.py
#
# Run this script with specified source_dir and target_dir while the model is training to make a copy
# of every checkpoint (checkpoints are kept once an hour by default and is difficult to adjust)
#
########

#%% Imports and constants

import time
import os
import shutil

check_every_n_minutes = 10

source_dir = '/datadrive/megadetectorv3/experiments/190425'
target_dir = '/datadrive/megadetectorv3/experiments/0425_checkpoints'

os.makedirs(target_dir, exist_ok=True)


#%% Main loop

num_checks = 0

while True:
    
    num_checks += 1
    print('Checking round {}.'.format(num_checks))

    for f in os.listdir(source_dir):
        # do not copy event or evaluation results
        if f.startswith('model') or f.startswith('graph'):
            target_path = os.path.join(target_dir, f)
            if not os.path.exists(target_path):
                _ = shutil.copy(os.path.join(source_dir, f), target_path)
                print('Copied {}.'.format(f))

    print('End of round {}.'.format(num_checks))

    time.sleep(check_every_n_minutes * 60)



"""

generate_crops_from_cct.py

Given a .json file in COCO Camera Traps format, creates a cropped image for
each bounding box.

"""

#%% Imports and constants

import os
import json

from tqdm import tqdm
from PIL import Image


#%% Functions

def generate_crops_from_cct(cct_file,image_dir,output_dir,padding=0,flat_output=True):
    """
    Given a .json file in COCO Camera Traps format, creates a cropped image for
    each bounding box.
    
    Args:
        cct_file (str): the COCO .json file from which we should load data
        image_dir (str): the folder where the images live; filenames in the .json
            file should be relative to this folder
        output_dir (str): the folder where we should write cropped images
        padding (float, optional): number of pixels we should expand each box before
            cropping
        flat_output (bool, optional): if False, folder structure will be preserved
            in the output, e.g. the image a/b/c/d.jpg will result in image files 
            in the output folder called, e.g., a/b/c/d_crop_000_id_12345.jpg.  If
            [flat_output] is True, the corresponding output image will be 
            a_b_c_d_crop_000_id_12345.jpg.            
    """
    
    ## Read and validate input
    
    assert os.path.isfile(cct_file)
    assert os.path.isdir(image_dir)
    os.makedirs(output_dir,exist_ok=True)

    with open(cct_file,'r') as f:
        d = json.load(f)
   
    
    ## Find annotations for each image
    
    from collections import defaultdict
    
    # This actually maps image IDs to annotations, but only to annotations
    # containing boxes
    image_id_to_boxes = defaultdict(list)
    
    n_boxes = 0
    
    for ann in d['annotations']:
        if 'bbox' in ann:
            image_id_to_boxes[ann['image_id']].append(ann)
            n_boxes += 1
            
    print('Found {} boxes in {} annotations for {} images'.format(
        n_boxes,len(d['annotations']),len(d['images'])))
    
    
    ## Generate crops
        
    # im = d['images'][0]
    for im in tqdm(d['images']):
        
        input_image_fn = os.path.join(os.path.join(image_dir,im['file_name']))
        assert os.path.isfile(input_image_fn), 'Could not find image {}'.format(input_image_fn)
        
        if im['id'] not in image_id_to_boxes:
            continue
        
        annotations_this_image = image_id_to_boxes[im['id']]
        
        # Load the image
        img = Image.open(input_image_fn)
        
        # Generate crops
        # i_ann = 0; ann = annotations_this_image[i_ann]
        for i_ann,ann in enumerate(annotations_this_image):
            
            # x/y/w/h, origin at the upper-left
            bbox = ann['bbox']
            
            xmin = bbox[0]
            ymin = bbox[1]
            xmax = xmin + bbox[2]
            ymax = ymin + bbox[3]
            
            xmin -= padding / 2
            ymin -= padding / 2
            xmax += padding / 2
            ymax += padding / 2
            
            xmin = max(xmin,0)
            ymin = max(ymin,0)
            xmax = min(xmax,img.width-1)
            ymax = min(ymax,img.height-1)
                        
            crop = img.crop(box=[xmin, ymin, xmax, ymax])
            
            output_fn = os.path.splitext(im['file_name'])[0].replace('\\','/')
            if flat_output:
                output_fn = output_fn.replace('/','_')
            output_fn = output_fn + '_crop' + str(i_ann).zfill(3) + '_id_' + ann['id']
            output_fn = output_fn + '.jpg'
                
            output_full_path = os.path.join(output_dir,output_fn)
            
            if not flat_output:
                os.makedirs(os.path.dirname(output_full_path),exist_ok=True)
                
            crop.save(output_full_path)
            
        # ...for each box
        
    # ...for each image
    
# ...generate_crops_from_cct()


#%% Interactive driver

if False:
    
    pass

    #%%
    
    cct_file = os.path.expanduser('~/data/noaa/noaa_estuary_fish.json')
    image_dir = os.path.expanduser('~/data/noaa/JPEGImages')
    padding = 50
    flat_output = True
    output_dir = '/home/user/tmp/noaa-fish-crops'
    
    generate_crops_from_cct(cct_file,image_dir,output_dir,padding,flat_output=True)
    files = os.listdir(output_dir)
    
    
#%% Command-line driver

# TODO



"""

 carrizo_shrubfree_2018.py

 Convert the .csv file provided for the Carrizo Mojave data set to a 
 COCO-camera-traps .json file

"""

#%% Constants and environment

import pandas as pd
import os
import json
import uuid
import time
import humanfriendly

from tqdm import tqdm
from PIL import Image
import numpy as np

from md_utils.path_utils import find_images

input_base = r'Z:\Shrub-free zone Carrizo 2018'
input_metadata_file = os.path.join(input_base,'shrub-free 2018.csv')

output_base = r'G:\carrizo-mojave'
output_json_file = os.path.join(output_base,'carrizo shrub-free 2018.json')

image_directory = input_base

load_width_and_height = False

assert(os.path.isdir(image_directory))

category_replacements = {'ammosphermophilus nelsoni':'ammospermophilus nelsoni'}

annotation_fields_to_copy = ['rep','photo.rep','timeblock','night.day','observations']


#%% Read source data

input_metadata = pd.read_csv(input_metadata_file)

# Original .csv file had superfluous spaces in column names
input_metadata = input_metadata.rename(columns=lambda x: x.strip())

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))

input_metadata['file'] = 1 + input_metadata.groupby('rep').cumcount()
input_metadata['file'] = input_metadata[['file', 'rep']].apply(lambda x: "{0}\IMG_{1}.JPG".format(x[1], str(x[0]).zfill(4)), axis=1)


#%% Map filenames to rows, verify image existence
    
start_time = time.time()
filenames_to_rows = {}
image_filenames = input_metadata.file

missing_files = []
duplicate_rows = []

# Build up a map from filenames to a list of rows, checking image existence as we go
for iFile, fn in tqdm(enumerate(image_filenames),total=len(image_filenames)):
    if (fn in filenames_to_rows):
        duplicate_rows.append(iFile)
        filenames_to_rows[fn].append(iFile)
    else:
        filenames_to_rows[fn] = [iFile]
        image_path = os.path.join(image_directory, fn)
        if not os.path.isfile(image_path):
            missing_files.append(fn)

elapsed = time.time() - start_time

print('Finished verifying image existence in {}, found {} missing files (of {})'.format(
    humanfriendly.format_timespan(elapsed), 
    len(missing_files),len(image_filenames)))

assert len(duplicate_rows) == 0

# 58 missing files (of 17652)


#%% Check for images that aren't included in the metadata file

image_full_paths = find_images(image_directory, bRecursive=True)
images_missing_from_metadata = []

for iImage, image_path in tqdm(enumerate(image_full_paths),total=len(image_full_paths)):

    relative_path = os.path.relpath(image_path,input_base)
    if relative_path not in filenames_to_rows:
        images_missing_from_metadata.append(relative_path)
    
print('{} of {} files are not in metadata'.format(len(images_missing_from_metadata),len(image_full_paths)))

# 3012 of 20606 files are not in metadata


#%% Create CCT dictionaries

images = []
annotations = []

# Map categories to integer IDs
#
# The category '0' is reserved for 'empty'

categories_to_category_id = {}
categories_to_counts = {}
categories_to_category_id['empty'] = 0
categories_to_counts['empty'] = 0

next_category_id = 1

# For each image
#
# Because in practice images are 1:1 with annotations in this data set,
# this is also a loop over annotations.

start_time = time.time()

for image_name in image_filenames:
    
    rows = filenames_to_rows[image_name]
    
    # Each filename should just match one row
    assert(len(rows) == 1)   
    
    iRow = rows[0]
    row = input_metadata.iloc[iRow]
    im = {}
    im['id'] = image_name.split('.')[0]
    im['file_name'] = image_name
    im['datetime'] = row['date']
    im['location'] = "{0}_{1}_{2}".format(row['region'], row['site'], row['microsite'])
          
    image_path = os.path.join(image_directory, image_name)
    
    # Don't include images that don't exist on disk
    if not os.path.isfile(image_path):
        continue
    
    if load_width_and_height:
        pilImage = Image.open(image_path)
        width, height = pilImage.size
        im['width'] = width
        im['height'] = height
    else:
        im['width'] = -1
        im['height'] = -1
        
    images.append(im)
    
    is_image = row['animal.capture']
        
    if (is_image == 0):
        category = 'empty'
    else:
        if row['latin.bionomial'] is np.nan:
            category = 'unidentifiable'
        else:
            category = row['latin.bionomial'].replace('  ',' ').lower().strip()
        
    if category in category_replacements:
        category = category_replacements[category]
    
    # Have we seen this category before?
    if category in categories_to_category_id:
        categoryID = categories_to_category_id[category]
        categories_to_counts[category] += 1
    else:
        categoryID = next_category_id
        categories_to_category_id[category] = categoryID
        categories_to_counts[category] = 1
        next_category_id += 1
    
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = categoryID
    
    for fieldname in annotation_fields_to_copy:
        ann[fieldname] = row[fieldname]
        if ann[fieldname] is np.nan:
            ann[fieldname] = ''
        ann[fieldname] = str(ann[fieldname])
    
    annotations.append(ann)
    
# ...for each image
       
# Convert categories to a CCT-style dictionary
categories = []

for category in categories_to_counts:
    print('Category {}, count {}'.format(category,categories_to_counts[category]))
    categoryID = categories_to_category_id[category]
    cat = {}
    cat['name'] = category
    cat['id'] = categoryID
    categories.append(cat)    
    
elapsed = time.time() - start_time
print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))
    

#%% Create info struct

info = {}
info['year'] = 2018
info['version'] = 1
info['description'] = 'Carrizo Shrub-Free 2018'
info['contributor'] = 'York University'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_json_file, 'w'), indent=4)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate output

from data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_directory
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False
data = integrity_check_json_db.integrity_check_json_db(output_json_file,options)


#%% Preview labels

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = None
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['empty']
html_output_file,image_db = visualize_db.visualize_db(db_path=output_json_file,
                                                        output_dir=os.path.join(
                                                        output_base, 'carrizo shrub-free 2018/preview'),
                                                        image_base_dir=image_directory,
                                                        options=viz_options)
os.startfile(html_output_file)




"""

string_utils.py

Miscellaneous string utilities.

"""

#%% Imports

import re


#%% Functions

def is_float(s):
    """ 
    Checks whether [s] is an object (typically a string) that can be cast to a float
    
    Args:
        s (object): object to evaluate
        
    Returns:
        bool: True if s successfully casts to a float, otherwise False
    """
    
    try:
        _ = float(s)
    except ValueError:
        return False
    return True


def human_readable_to_bytes(size):
    """
    Given a human-readable byte string (e.g. 2G, 10GB, 30MB, 20KB),
    returns the number of bytes.  Will return 0 if the argument has
    unexpected form.
    
    https://gist.github.com/beugley/ccd69945346759eb6142272a6d69b4e0
    
    Args:
        size (str): string representing a size
        
    Returns:
        int: the corresponding size in bytes
    """
    
    size = re.sub(r'\s+', '', size)
    
    if (size[-1] == 'B'):
        size = size[:-1]
        
    if (size.isdigit()):
        bytes = int(size)
    elif (is_float(size)):
        bytes = float(size)
    else:
        bytes = size[:-1]
        unit = size[-1]
        try:        
            bytes = float(bytes)
            if (unit == 'T'):
                bytes *= 1024*1024*1024*1024
            elif (unit == 'G'):
                bytes *= 1024*1024*1024
            elif (unit == 'M'):
                bytes *= 1024*1024
            elif (unit == 'K'):
                bytes *= 1024
            else:
                bytes = 0
        except ValueError:
            bytes = 0
            
    return bytes


def remove_ansi_codes(s):
    """
    Removes ANSI escape codes from a string.
    
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python#14693789
    
    Args:
        s (str): the string to de-ANSI-i-fy
        
    Returns:
        str: A copy of [s] without ANSI codes
    """
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', s)



"""

retrieve_sample_image.py

Downloader that retrieves images from Google images, used for verifying taxonomy
lookups and looking for egregious mismappings (e.g., "snake" being mapped to a fish called
"snake").

Simple wrapper around simple_image_download, but I've had to swap in and out the underlying
downloader a few times.

"""

#%% Imports and environment

import os

output_folder = os.path.expanduser('~/tmp/image-download-test')
os.makedirs(output_folder,exist_ok=True)
                                   
method = 'simple_image_download' # 'google_images_download'

if method == 'simple_image_download':
    
    from taxonomy_mapping import simple_image_download
    google_image_downloader = simple_image_download.Downloader()
    google_image_downloader.directory = output_folder
 
elif method == 'google_images_download':
    
    from google_images_download import google_images_download

else:
    
    raise ValueError('Unrecognized method {}'.format(method))


#%% Main entry point

def download_images(query,output_directory,limit=100,verbose=False):

    query = query.replace(' ','+')        
    
    if method == 'simple_image_download':
        
        google_image_downloader.directory = output_directory
        paths = google_image_downloader.download(query, limit=limit,
          verbose=verbose, cache=False, download_cache=False)
        return paths
        
    elif method == 'google_images_download':
        
        response = google_images_download.googleimagesdownload()    
        arguments = {'keywords':query,'limit':limit,'print_urls':verbose,
                     'image-directory':output_directory}
        response.download(arguments)
        return None

    else:
        
        raise ValueError('Unrecognized method {}'.format(method))
        

#%% Test driver

if False:
    
    #%%
    
    paths = download_images(query='redunca',output_directory=output_folder,
                    limit=20,verbose=True) 



"""

 sulross_get_exif.py

 For the Sul Ross dataset, species informationw was stored in XMP metadata; pull 
 all that metadata out to .json.

"""

import os
import json
from tqdm import tqdm

import exiftool

image_ids_path = '/home/beaver/cameratraps/data/sulross/20190522_image_ids.json'
data_dir = '/home/beaver/cameratraps/mnt/sulross'


def get_metadata():
    
    image_ids = json.load(open(image_ids_path))

    image_id_to_metadata = {}

    # exiftool can process a batch of images at a time, but bottleneck is blobfuse reading the images
    batch_size = 20  

    num_images_processed = 0

    with exiftool.ExifTool() as et:
        for i in tqdm(range(0, len(image_ids), batch_size)):
            batch_ids = image_ids[i: i + batch_size]

            batch_paths = [os.path.join(data_dir, i) for i in batch_ids]

            try:
                metadatas = et.get_metadata_batch(batch_paths)

                for id, metadata in zip(batch_ids, metadatas):
                    image_id_to_metadata[id] = metadata['XMP:HierarchicalSubject']
            except Exception as e:
                print('Exception! {}'.format(e))
                continue

            num_images_processed += batch_size
            if num_images_processed % 1000 == 0:
                print('Finished processing {} images; image ID {}'.format(
                    num_images_processed, image_ids[num_images_processed - 1]))
                print(image_id_to_metadata[id])
                print()

            # checkpoint
            if num_images_processed % 10000 == 0:
                print('Saving results so far...')
                with open('/home/beaver/cameratraps/data/sulross/20190522_metadata.json', 'w') as f:
                    json.dump(image_id_to_metadata, f, indent=1)

    print('Length of meta data read: ', len(image_id_to_metadata))
    with open('/home/beaver/cameratraps/data/sulross/20190522_metadata.json', 'w') as f:
        json.dump(image_id_to_metadata, f, indent=1)
    print('Results saved. Done!')

if __name__ == '__main__':
    get_metadata()



"""

get_image_sizes.py

Given a json-formatted list of image filenames, retrieves the width and height of 
every image, optionally writing the results to a new .json file.

"""

#%% Constants and imports

import argparse
import json
import os
from PIL import Image
import sys

from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from functools import partial
from tqdm import tqdm

from md_utils.path_utils import find_images

image_base = ''
default_n_threads = 1
use_threads = False


#%% Processing functions

def _get_image_size(image_path,image_prefix=None):
    """
    Support function to get the size of a single image.  Returns a (path,w,h) tuple.
    w and h will be -1 if the image fails to load.
    """
    
    if image_prefix is not None:
        full_path = os.path.join(image_prefix,image_path)
    else:
        full_path = image_path
    
    # Is this image on disk?
    if not os.path.isfile(full_path):
        print('Could not find image {}'.format(full_path))
        return (image_path,-1,-1)

    try:        
        pil_im = Image.open(full_path)
        w = pil_im.width            
        h = pil_im.height
        return (image_path,w,h)
    except Exception as e:    
        print('Error reading image {}: {}'.format(full_path,str(e)))
        return (image_path,-1,-1)
    
    
def get_image_sizes(filenames,image_prefix=None,output_file=None,
                    n_workers=default_n_threads,use_threads=True,
                    recursive=True):
    """
    Gets the width and height of all images in [filenames], which can be:
        
    * A .json-formatted file containing list of strings
    * A folder
    * A list of files

    ...returning a list of (path,w,h) tuples, and optionally writing the results to [output_file].
    
    Args:
        filenames (str or list): the image filenames for which we should retrieve sizes, 
            can be the name of a .json-formatted file containing list of strings, a folder 
            in which we should enumerate images, or a list of files.
        image_prefix (str, optional): optional prefix to add to images to get to full paths;
            useful when [filenames] contains relative files, in which case [image_prefix] is the 
            base folder for the source images.
        output_file (str, optional): a .json file to write the imgae sizes
        n_workers (int, optional): number of parallel workers to use, set to <=1 to
            disable parallelization
        use_threads (bool, optional): whether to use threads (True) or processes (False)
            for parallelization; not relevant if [n_workers] <= 1
        recursive (bool, optional): only relevant if [filenames] is actually a folder,
            determines whether image enumeration within that folder will be recursive
        
    Returns:
        list: list of (path,w,h) tuples
    """        
    
    if output_file is not None:
        assert os.path.isdir(os.path.dirname(output_file)), \
            'Illegal output file {}, parent folder does not exist'.format(output_file)
        
    if isinstance(filenames,str) and os.path.isfile(filenames):
        with open(filenames,'r') as f:        
            filenames = json.load(f)
        filenames = [s.strip() for s in filenames]
    elif isinstance(filenames,str) and os.path.isdir(filenames):
        filenames = find_images(filenames,recursive=recursive,
                                return_relative_paths=False,convert_slashes=True)
    else:
        assert isinstance(filenames,list)        
    
    if n_workers <= 1:
        
        all_results = []
        for i_file,fn in tqdm(enumerate(filenames),total=len(filenames)):
            all_results.append(_get_image_size(fn,image_prefix=image_prefix))
    
    else:
        
        print('Creating a pool with {} workers'.format(n_workers))
        if use_threads:
            pool = ThreadPool(n_workers)        
        else:
            pool = Pool(n_workers)
        # all_results = list(tqdm(pool.imap(process_image, filenames), total=len(filenames)))
        all_results = list(tqdm(pool.imap(
            partial(_get_image_size,image_prefix=image_prefix), filenames), total=len(filenames)))
    
    if output_file is not None:
        with open(output_file,'w') as f:
            json.dump(all_results,f,indent=1)
            
    return all_results

    
#%% Interactive driver

if False:

    pass    

    #%%
    
    # List images in a test folder
    base_dir = r'c:\temp\test_images'
    image_list_file = os.path.join(base_dir,'images.json')
    relative_image_list_file = os.path.join(base_dir,'images_relative.json')
    image_size_file = os.path.join(base_dir,'image_sizes.json')
    from md_utils import path_utils
    image_names = path_utils.find_images(base_dir,recursive=True)
    
    with open(image_list_file,'w') as f:
        json.dump(image_names,f,indent=1)
        
    relative_image_names = []
    for s in image_names:
        relative_image_names.append(os.path.relpath(s,base_dir))
    
    with open(relative_image_list_file,'w') as f:
        json.dump(relative_image_names,f,indent=1)
    
    
    #%%
    
    get_image_sizes(relative_image_list_file,image_size_file,image_prefix=base_dir,n_threads=4)
    
    
#%% Command-line driver
    
def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('filenames',type=str,
                        help='Folder from which we should fetch image sizes, or .json file with a list of filenames')
    parser.add_argument('output_file',type=str,
                        help='Output file (.json) to which we should write image size information')
    parser.add_argument('--image_prefix', type=str, default=None,
                        help='Prefix to append to image filenames, only relevant if [filenames] points to a list of ' + \
                             'relative paths')        
    parser.add_argument('--n_threads', type=int, default=default_n_threads,
                        help='Number of concurrent workers, set to <=1 to disable parallelization (default {})'.format(
                            default_n_threads))
                        
    if len(sys.argv[1:])==0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()
    
    _ = get_image_sizes(filenames=args.filenames,
                        output_file=args.output_file,
                        image_prefix=args.image_prefix,
                        n_workers=args.n_threads)
        
if __name__ == '__main__':
    
    main()



########
#
# annotation_constants.py
#
# Shared constants used to interpret annotation output
#
# Categories assigned to bounding boxes.  Used throughout our repo; do not change unless
# you are Dan or Siyu.  In fact, do not change unless you are both Dan *and* Siyu.
# 
# We use integer indices here; this is different than the API output .json file,
# where indices are string integers.
#
########

NUM_DETECTOR_CATEGORIES = 3  # this is for choosing colors, so ignoring the "empty" class

# This is the label mapping used for our incoming iMerit annotations
# Only used to parse the incoming annotations. In our database, the string name is used to avoid confusion
annotation_bbox_categories = [
    {'id': 0, 'name': 'empty'},
    {'id': 1, 'name': 'animal'},
    {'id': 2, 'name': 'person'},
    {'id': 3, 'name': 'group'},  # group of animals
    {'id': 4, 'name': 'vehicle'}
]

annotation_bbox_category_id_to_name = {}
annotation_bbox_category_name_to_id = {}

for cat in annotation_bbox_categories:
    annotation_bbox_category_id_to_name[cat['id']] = cat['name']
    annotation_bbox_category_name_to_id[cat['name']] = cat['id']

# MegaDetector outputs
detector_bbox_categories = [
    {'id': 0, 'name': 'empty'},
    {'id': 1, 'name': 'animal'},
    {'id': 2, 'name': 'person'},
    {'id': 3, 'name': 'vehicle'}
]

detector_bbox_category_id_to_name = {}
detector_bbox_category_name_to_id = {}

for cat in detector_bbox_categories:
    detector_bbox_category_id_to_name[cat['id']] = cat['name']
    detector_bbox_category_name_to_id[cat['name']] = cat['id']
